import logging

from django.db import migrations
from django.db.models import Q

logger = logging.getLogger(__name__)


def update_subscription_criteria(apps, schema_editor):
    resource_type_map = {
        'Invoice': 'Invoice',
        'Bill': 'Invoice',
        'Patient': 'Patient',
        'Organisation': 'Organisation'
    }

    subscription_model = apps.get_model('api_fhir_r4', 'Subscription')
    entries_to_update = subscription_model.objects\
        .filter(Q(criteria__has_key='resource_type')  & ~Q(criteria__has_key='resource'))
    for subscription in entries_to_update:
        resource = subscription.criteria.pop('resource_type')
        if resource in resource_type_map:
            subscription.criteria['resource'] = resource_type_map[resource]
            subscription.save()
        else:
            logger.warning(f'Subscription of unexpected type - {resource}')


class Migration(migrations.Migration):

    dependencies = [
        ('api_fhir_r4', '0003_auto_20220313_1634'),
    ]

    operations = [
        migrations.RunPython(update_subscription_criteria, migrations.RunPython.noop)
    ]



from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter, EnrolmentOfficerPractitionerConverter, ReferenceConverterMixin
from api_fhir_r4.converters.healthFacilityOrganisationConverter import LocationConverter, PersonConverterMixin
from api_fhir_r4.utils import DbManagerUtils
from core.models import Officer
from django.utils.translation import gettext as _
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.practitionerrole import PractitionerRole


class EnrolmentOfficerPractitionerRoleConverter(BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_officer, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_practitioner_role = PractitionerRole.construct()
        cls.build_fhir_pk(fhir_practitioner_role, imis_officer, reference_type)
        cls.build_fhir_identifiers(fhir_practitioner_role, imis_officer)
        cls.build_fhir_extension(fhir_practitioner_role, imis_officer, reference_type)
        cls.build_fhir_practitioner_reference(fhir_practitioner_role, imis_officer, reference_type)
        cls.build_fhir_location_references(fhir_practitioner_role, imis_officer, reference_type)
        cls.build_fhir_code(fhir_practitioner_role)
        cls.build_fhir_telecom(fhir_practitioner_role, imis_officer)
        return fhir_practitioner_role

    @classmethod
    def to_imis_obj(cls, fhir_practitioner_role, audit_user_id):
        errors = []
        fhir_practitioner_role = PractitionerRole(**fhir_practitioner_role)
        practitioner = fhir_practitioner_role.practitioner
        imis_officer = EnrolmentOfficerPractitionerConverter.get_imis_obj_by_fhir_reference(practitioner, errors)
        location_references = fhir_practitioner_role.location
        if len(location_references)!=1:
            errors.append("Location does not have 1 element")
        else:
            location = cls.get_location_by_reference(location_references[0], errors)
            substitution_officer = None
        if fhir_practitioner_role.extension and len(fhir_practitioner_role.extension) > 0:
            substitution_officer_reference = fhir_practitioner_role.extension[0].valueReference
            substitution_officer = EnrolmentOfficerPractitionerConverter.get_imis_obj_by_fhir_reference(
                substitution_officer_reference,
            )
        if not cls.valid_condition(imis_officer is None, "Practitioner doesn't exists", errors):
            imis_officer.location = location
            if substitution_officer:
                imis_officer.substitution_officer = substitution_officer
        cls.check_errors(errors)
        return imis_officer

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_reference_obj_uuid(cls, imis_officer):
        return imis_officer.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_officer):
        return imis_officer.id

    @classmethod
    def get_reference_obj_code(cls, imis_officer):
        return imis_officer.code

    @classmethod
    def get_fhir_resource_type(cls):
        return PractitionerRole

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        return DbManagerUtils.get_object_or_none(
            Officer,
            **cls.get_database_query_id_parameteres_from_reference(reference))

        imis_officer_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Officer, code=imis_officer_code)
        return DbManagerUtils.get_object_or_none(
            Officer,
            **cls.get_database_query_id_parameteres_from_reference(reference))


    @classmethod
    def build_fhir_extension(cls, fhir_practitioner_role, imis_officer, reference_type):
        if imis_officer.substitution_officer:
            extension = Extension.construct()
            extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/practitioner-role-substitution-reference"

            reference = EnrolmentOfficerPractitionerConverter.build_fhir_resource_reference(
                imis_officer.substitution_officer,
                reference_type=reference_type
            )
            extension.valueReference = reference
            fhir_practitioner_role.extension = [extension]

    @classmethod
    def build_fhir_identifiers(cls, fhir_practitioner_role, imis_officer):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_officer)
        fhir_practitioner_role.identifier = identifiers

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_officer):
        if imis_officer.code:
            identifier = cls.build_fhir_identifier(imis_officer.code,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_generic_type_code())
            identifiers.append(identifier)

    @classmethod
    def build_fhir_practitioner_reference(cls, fhir_practitioner_role, imis_officer, reference_type):
        fhir_practitioner_role.practitioner = EnrolmentOfficerPractitionerConverter\
            .build_fhir_resource_reference(imis_officer, reference_type=reference_type)

    @classmethod
    def build_fhir_location_references(cls, fhir_practitioner_role, imis_officer, reference_type):
        if imis_officer.location:
            reference = LocationConverter.build_fhir_resource_reference(
                imis_officer.location, reference_type=reference_type)
            fhir_practitioner_role.location = [reference]

    @classmethod
    def get_location_by_reference(cls, location_references, errors):
        return LocationConverter.get_imis_obj_by_fhir_reference(location_references)

    @classmethod
    def get_first_location(cls, location_references):
        return location_references[0]

    @classmethod
    def build_fhir_code(cls, fhir_practitioner_role):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/practitioner-qualification-type"
        fhir_practitioner_role.code = [cls.build_codeable_concept(
            system=system,
            code="EO",
            display=_("Enrolment Officer")
        )]

    @classmethod
    def build_fhir_telecom(cls, fhir_practitioner_role, imis_officer):
        fhir_practitioner_role.telecom = cls.build_fhir_telecom_for_person(
            phone=imis_officer.phone,
            email=imis_officer.email
        )



from api_fhir_r4.converters import OperationOutcomeConverter
from api_fhir_r4.models import OperationOutcomeV2
from api_fhir_r4.tests import OperationOutcomeTestMixin
from api_fhir_r4.tests.mixin import ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class OperationOutcomeConverterTestCase(OperationOutcomeTestMixin,
                                        ConvertToFhirTestMixin,
                                        ConvertJsonToFhirTestMixin):
    converter = OperationOutcomeConverter
    fhir_resource = OperationOutcomeV2
    json_repr = 'test/test_outcome.json'



import json
import os

from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.tests import LocationTestMixin, ClaimAdminPractitionerTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin
from api_fhir_r4.utils import TimeUtils
from api_fhir_r4.tests.utils import load_and_replace_json,get_connection_payload,get_or_create_user_api
from api_fhir_r4.utils import DbManagerUtils

from claim.models import Claim, ClaimDetail
from insuree.test_helpers import create_test_insuree
from location.models import HealthFacility, UserDistrict
from location.test_helpers import create_test_village, create_test_health_facility
from medical.models import Diagnosis
from medical.test_helpers import create_test_item, create_test_service
from claim.test_helpers import create_test_claimservice,create_test_claimitem,create_test_claim_admin

class ClaimAPITests(GenericFhirAPITestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url() + 'Claim/'
    _test_json_path = "/test/test_claim.json"

    # diagnosis data
    _TEST_MAIN_ICD_CODE = 'T_CD'
    _TEST_MAIN_ICD_NAME = 'Test diagnosis'

    _TEST_CLAIM_ADMIN_UUID = "044c33d1-dbf3-4d6a-9924-3797b461e535"
    _TEST_CLAIM_ADMIN_CODE = "VITEST99"
    _TEST_INSUREE_CODE = "999000001"
    _TEST_INSUREE_UUID = '76aca309-f8cf-4890-8f2e-b416d78de00b'
    # claim item data
    _TEST_ITEM_CODE = "0004"
    _TEST_ITEM_UUID = "e2bc1546-390b-4d41-8571-632ecf7a0936"
    _TEST_ITEM_QUANTITY_PROVIDED = 10.0
    _TEST_ITEM_PRICE_ASKED = 10.0
    _TEST_ITEM_EXPLANATION = "item_explanation"
    _TEST_ITEM_TYPE = 'D'

    # claim service data
    _TEST_SERVICE_CODE = "M7"
    _TEST_SERVICE_UUID = "a17602f4-e9ff-4f42-a6a4-ccefcb23b4d6"
    _TEST_SERVICE_QUANTITY_PROVIDED = 1
    _TEST_SERVICE_PRICE_ASKED = 21000.0
    _TEST_SERVICE_EXPLANATION = "service_explanation"
    _TEST_SERVICE_TYPE = 'D'

    # hf test data
    _TEST_HF_ID = 10000
    _TEST_HF_UUID = "6d0eea8c-62eb-11ea-94d6-c36229a16c2f"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"

    _ADMIN_AUDIT_USER_ID = -1
    _TEST_USER = None
    _TEST_USER_NAME = "TestUserTest2"
    _TEST_USER_PASSWORD = "TestPasswordTest2"
    _TEST_DATA_USER = {
        "username": _TEST_USER_NAME,
        "last_name": _TEST_USER_NAME,
        "password": _TEST_USER_PASSWORD,
        "other_names": _TEST_USER_NAME,
        "user_types": "INTERACTIVE",
        "language": "en",
        "roles": [9],
    }

    _test_json_path_credentials = "/test/test_login.json"
    _test_request_data_credentials = None
    _test_json_path_with_code_references = "/test/test_claim_with_code_references.json"
    sub_str = {}
    test_insuree = None
    test_claim_admin = None
 
    def setUp(self):
        super(ClaimAPITests, self).setUp()
        self._TEST_USER = get_or_create_user_api(self._TEST_DATA_USER)
        self.create_dependencies()
        self.sub_str[self._TEST_INSUREE_UUID]=self.test_insuree.uuid
        self.sub_str[self._TEST_INSUREE_CODE]=self.test_insuree.chf_id
        self.sub_str[self._TEST_HF_UUID]=self.test_hf.uuid
        self.sub_str[self._TEST_CLAIM_ADMIN_UUID]=self.test_claim_admin.uuid
        self.sub_str[self._TEST_CLAIM_ADMIN_CODE]=self.test_claim_admin.code

        self.sub_str[self._TEST_ITEM_UUID]=self._TEST_ITEM.uuid
        self.sub_str[self._TEST_SERVICE_UUID]=self._TEST_SERVICE.uuid
        self.sub_str[self._TEST_ITEM_CODE]=self._TEST_ITEM.code
        self.sub_str[self._TEST_SERVICE_CODE]=self._TEST_SERVICE.code
        self._TEST_HF_ID = self.test_hf.id
        self._TEST_HF_UUID = self.test_hf.uuid



    def create_dependencies(self):
        self.test_icd = Diagnosis(code=self._TEST_MAIN_ICD_CODE, name =self._TEST_MAIN_ICD_NAME)
        self.test_icd.audit_user_id = self._ADMIN_AUDIT_USER_ID
        self.test_icd.save()
        self.test_insuree = create_test_insuree(custom_props={"chf_id": self._TEST_INSUREE_CODE})
        self.test_hf = self.create_test_hf()
        if not self.test_claim_admin:
            self.test_claim_admin =create_test_claim_admin(
                custom_props={'code':'T-CA-API',
                              'health_facility_id': self.test_hf.id,
                                'last_name' : self._TEST_DATA_USER['last_name'],
                                'other_names' : self._TEST_DATA_USER['other_names']})
        
        self._TEST_USER.claim_admin = self.test_claim_admin
        self._TEST_USER.save()
        ud = UserDistrict()
        ud.location = self.test_insuree.family.location.parent.parent
        ud.audit_user_id = self._ADMIN_AUDIT_USER_ID
        ud.user = self._TEST_USER.i_user
        ud.validity_from = TimeUtils.now()
        ud.save()

        self._TEST_ITEM = create_test_item(self._TEST_SERVICE_TYPE)
        self._TEST_SERVICE = create_test_service(self._TEST_ITEM_TYPE)

    def create_test_hf(self):
        self.test_hf = create_test_health_facility(
            self._TEST_HF_CODE,
            self.test_insuree.family.location.parent.parent.id,
            custom_props = {
                'name': self._TEST_HF_NAME,
                'level':self._TEST_HF_LEVEL,
                'legal_form_id':self._TEST_HF_LEGAL_FORM,
                'address':self._TEST_ADDRESS,
                'phone':self._TEST_PHONE,
                'fax':self._TEST_FAX,
                'email':self._TEST_EMAIL,
            }
        )
        return self.test_hf



    def _post_claim(self, data, headers):
        return self.client.post(self.base_url, data=data, format='json', **headers)


    def test_post_should_create_correctly(self):
        response = self.client.post(
            GeneralConfiguration.get_base_url() + 'login/', data=get_connection_payload(self._TEST_DATA_USER), format='json'
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}"
        }

        dataset = [
            load_and_replace_json(self._test_json_path,self.sub_str),
            load_and_replace_json(self._test_json_path_with_code_references,self.sub_str)
        ]
        for data in dataset:
            response = self.client.post(self.base_url, data=data, format='json', **headers)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
            self.assertIsNotNone(response.content)
            response_json = response.json()
            self.assertEqual(response_json["resourceType"], 'ClaimResponse')
            self.assertEqual(response_json["outcome"], 'complete')
            # for both tests item and service should be 'rejected' and 'not in price list'
            for item in response_json["item"]:
                for adjudication in item["adjudication"]:
                    self.assertEqual(adjudication["category"]["coding"][0]["code"], f'{Claim.STATUS_REJECTED}')
                    # 2 not in price list
                    self.assertEqual(adjudication["reason"]["coding"][0]["code"], '2')


    def test_get_should_return_200_claim_response(self):
        # test if get ClaimResponse return 200
        response = self.client.post(
            GeneralConfiguration.get_base_url() + 'login/', data=get_connection_payload(self._TEST_DATA_USER), format='json'
        )
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            "HTTP_AUTHORIZATION": f"Bearer {token}"
        }
        response = self.client.get(GeneralConfiguration.get_base_url() + 'ClaimResponse/', data=None, format='json',
                                   **headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK)



import core
import decimal

from product.models import Product

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import InsurancePlanConverter
from api_fhir_r4.tests import GenericTestMixin

from django.utils.translation import gettext as _
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.money import Money
from fhir.resources.R4B.insuranceplan import InsurancePlan, InsurancePlanCoverage, \
    InsurancePlanCoverageBenefit, InsurancePlanCoverageBenefitLimit, \
    InsurancePlanPlan, InsurancePlanPlanGeneralCost
from fhir.resources.R4B.period import Period
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.quantity import Quantity
from api_fhir_r4.utils import TimeUtils


class InsurancePlanTestMixin(GenericTestMixin):

    _TEST_PRODUCT_CODE = "TEST0001"
    _TEST_PRODUCT_NAME = "Test Product"
    _TEST_PERIOD_START = "2017-01-01T00:00:00"
    _TEST_PERIOD_END = "2030-12-31T00:00:00"
    _TEST_MAX_INSTALLMENTS = 3
    _TEST_GRACE_PERIOD_PAYMENT = 1
    _TEST_INSURANCE_PERIOD = 12
    _TEST_MEMBER_COUNT = 9999
    _TEST_LUMPSUM = 0
    _TEST_THRESHOLD = 6
    _TEST_PREMIUM_ADULT = 4000
    _TEST_RENEWAL_DISCOUNT = 40
    _TEST_RENEWAL_DISCOUNT_PERIOD = 1
    _TEST_ENROLMENT_DISCOUNT = 30
    _TEST_ENROLMENT_DISCOUNT_PERIOD = 1

    def create_test_imis_instance(self):
        imis_product = Product()
        imis_product.code = self._TEST_PRODUCT_CODE
        imis_product.name = self._TEST_PRODUCT_NAME
        imis_product.date_from = TimeUtils.str_to_date(self._TEST_PERIOD_START)
        imis_product.date_to = TimeUtils.str_to_date(self._TEST_PERIOD_END)
        imis_product.max_installments = self._TEST_MAX_INSTALLMENTS
        imis_product.grace_period_enrolment = self._TEST_GRACE_PERIOD_PAYMENT
        imis_product.insurance_period = self._TEST_INSURANCE_PERIOD
        imis_product.max_members = self._TEST_MEMBER_COUNT
        imis_product.lump_sum = self._TEST_LUMPSUM
        imis_product.threshold = self._TEST_THRESHOLD
        imis_product.premium_adult = self._TEST_PREMIUM_ADULT
        imis_product.renewal_discount_perc = self._TEST_RENEWAL_DISCOUNT
        imis_product.renewal_discount_period = self._TEST_RENEWAL_DISCOUNT_PERIOD
        imis_product.enrolment_discount_perc = self._TEST_ENROLMENT_DISCOUNT
        imis_product.enrolment_discount_period = self._TEST_ENROLMENT_DISCOUNT_PERIOD
        return imis_product

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_PRODUCT_CODE, imis_obj.code)
        self.assertEqual(self._TEST_PRODUCT_NAME, imis_obj.name)
        self.assertEqual(self._TEST_PERIOD_START, imis_obj.date_from.isoformat())
        self.assertEqual(self._TEST_PERIOD_END, imis_obj.date_to.isoformat())
        self.assertEqual(self._TEST_MAX_INSTALLMENTS, imis_obj.max_installments)
        self.assertEqual(self._TEST_GRACE_PERIOD_PAYMENT, imis_obj.grace_period_enrolment)
        self.assertEqual(self._TEST_INSURANCE_PERIOD, imis_obj.insurance_period)
        self.assertEqual(self._TEST_MEMBER_COUNT, imis_obj.max_members)
        self.assertEqual(self._TEST_LUMPSUM, imis_obj.lump_sum)
        self.assertEqual(self._TEST_THRESHOLD, imis_obj.threshold)
        self.assertEqual(self._TEST_PREMIUM_ADULT, imis_obj.premium_adult)

    def create_test_fhir_instance(self):
        if hasattr(core, 'currency'):
            currency = core.currency
        else:
            currency = "EUR"

        fhir_insurance_plan = InsurancePlan.construct()

        code = InsurancePlanConverter.build_fhir_identifier(
            self._TEST_PRODUCT_CODE,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        identifiers = [code]
        fhir_insurance_plan.identifier = identifiers
        fhir_insurance_plan.name = self._TEST_PRODUCT_NAME

        type = InsurancePlanConverter.build_codeable_concept(
            code="medical",
            system="http://terminology.hl7.org/CodeSystem/insurance-plan-type"
        )
        if len(type.coding) == 1:
            type.coding[0].display = _("Medical")
        fhir_insurance_plan.type = [type]

        period = Period.construct()
        period.start = self._TEST_PERIOD_START
        period.end = self._TEST_PERIOD_END
        fhir_insurance_plan.period = period

        coverage = InsurancePlanCoverage.construct()
        coverage.type = InsurancePlanConverter.build_codeable_concept(
            code="medical",
            system="http://terminology.hl7.org/CodeSystem/insurance-plan-type"
        )

        # build coverage benefit
        benefit = InsurancePlanCoverageBenefit.construct()
        benefit.type = InsurancePlanConverter.build_codeable_concept(
            code="medical",
            system="http://terminology.hl7.org/CodeSystem/insurance-plan-type"
        )

        # build coverage benefit limit slices
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/insurance-plan-coverage-benefit-limit"
        limit = InsurancePlanCoverageBenefitLimit.construct()
        quantity = Quantity.construct()
        quantity.value = self._TEST_INSURANCE_PERIOD
        quantity.unit = "month"
        limit.value = quantity
        limit.code = InsurancePlanConverter.build_codeable_concept("period", system)
        limit.code.coding[0].display = _("Period")
        benefit.limit = [limit]

        limit = InsurancePlanCoverageBenefitLimit.construct()
        quantity = Quantity.construct()
        quantity.value = self._TEST_MEMBER_COUNT
        quantity.unit = "member"
        limit.value = quantity
        limit.code = InsurancePlanConverter.build_codeable_concept("memberCount", system)
        limit.code.coding[0].display = _("Member Count")
        benefit.limit.append(limit)

        coverage.benefit = [benefit]
        fhir_insurance_plan.coverage = [coverage]

        plan = InsurancePlanPlan.construct()
        # build plan general cost limit slices
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/insurance-plan-general-cost-type"
        general_cost = InsurancePlanPlanGeneralCost.construct()
        cost = Money.construct()
        cost.value = self._TEST_LUMPSUM
        cost.currency = currency
        general_cost.cost = cost
        general_cost.type = InsurancePlanConverter.build_codeable_concept("lumpsum", system)
        general_cost.type.coding[0].display = _("Lumpsum")
        plan.generalCost = [general_cost]
        plan.generalCost[0].groupSize = self._TEST_THRESHOLD

        general_cost = InsurancePlanPlanGeneralCost.construct()
        cost = Money.construct()
        cost.value = self._TEST_PREMIUM_ADULT
        cost.currency = currency
        general_cost.cost = cost
        general_cost.type = InsurancePlanConverter.build_codeable_concept("premiumAdult", system)
        general_cost.type.coding[0].display = _("Premium Adult")
        plan.generalCost.append(general_cost)

        fhir_insurance_plan.plan = [plan]

        # extensions
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-max-installments"
        extension.valueUnsignedInt = self._TEST_MAX_INSTALLMENTS
        fhir_insurance_plan.extension = [extension]
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-period"
        extension.valueQuantity = Quantity(
            **{
                "value": self._TEST_GRACE_PERIOD_PAYMENT,
                "unit": "months"
            }
        )
        fhir_insurance_plan.extension.append(extension)

        # discount processing - renewal
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-discount"
        nested_extension = Extension.construct()

        # percentage
        nested_extension.url = "Percentage"
        nested_extension.valueDecimal = self._TEST_RENEWAL_DISCOUNT
        extension.extension = [nested_extension]

        # period
        nested_extension = Extension.construct()
        nested_extension.url = "Period"
        nested_extension.valueQuantity = Quantity(
            **{
                "value": self._TEST_RENEWAL_DISCOUNT_PERIOD,
                "unit": "months"
            }
        )
        extension.extension.append(nested_extension)
        fhir_insurance_plan.extension.append(extension)

        # discount processing - enrolment
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-discount"
        nested_extension = Extension.construct()

        # percentage
        nested_extension.url = "Percentage"
        nested_extension.valueDecimal = self._TEST_ENROLMENT_DISCOUNT
        extension.extension = [nested_extension]

        # period
        nested_extension = Extension.construct()
        nested_extension.url = "Period"
        nested_extension.valueQuantity = Quantity(
            **{
                "value": self._TEST_ENROLMENT_DISCOUNT_PERIOD,
                "unit": "months"
            }
        )
        extension.extension.append(nested_extension)
        fhir_insurance_plan.extension.append(extension)

        return fhir_insurance_plan

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(4, len(fhir_obj.extension))
        extension_max_installment = fhir_obj.extension[0]
        self.assertTrue(isinstance(extension_max_installment, Extension))
        self.assertEqual(self._TEST_MAX_INSTALLMENTS, extension_max_installment.valueUnsignedInt)
        extension_grace_period = fhir_obj.extension[1].valueQuantity
        self.assertTrue(isinstance(extension_grace_period, Quantity))
        self.assertEqual(self._TEST_GRACE_PERIOD_PAYMENT, extension_grace_period.value)
        extension_renewal_discounts = fhir_obj.extension[2].extension
        self.assertAlmostEqual(decimal.Decimal(self._TEST_RENEWAL_DISCOUNT),
                               extension_renewal_discounts[0].valueDecimal, places=2)
        self.assertEqual(self._TEST_RENEWAL_DISCOUNT_PERIOD, extension_renewal_discounts[1].valueQuantity.value)
        extension_enrolment_discounts = fhir_obj.extension[3].extension
        self.assertAlmostEqual(decimal.Decimal(self._TEST_ENROLMENT_DISCOUNT),
                               extension_enrolment_discounts[0].valueDecimal, places=2)
        self.assertEqual(self._TEST_ENROLMENT_DISCOUNT_PERIOD, extension_enrolment_discounts[1].valueQuantity.value)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = InsurancePlanConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_claim_admin_code_type():
                self.assertEqual(self._TEST_PRODUCT_CODE, identifier.value)
        benefit_limit_period = fhir_obj.coverage[0].benefit[0].limit[0].value
        self.assertEqual(self._TEST_INSURANCE_PERIOD, benefit_limit_period.value)
        benefit_limit_member_count = fhir_obj.coverage[0].benefit[0].limit[1].value
        self.assertEqual(self._TEST_MEMBER_COUNT, benefit_limit_member_count.value)
        lumpsum = fhir_obj.plan[0].generalCost[0].cost.value
        self.assertEqual(self._TEST_LUMPSUM, lumpsum)
        threshold = fhir_obj.plan[0].generalCost[0].groupSize
        self.assertEqual(self._TEST_THRESHOLD, threshold)
        premium_adult = fhir_obj.plan[0].generalCost[1].cost.value
        self.assertEqual(self._TEST_PREMIUM_ADULT, premium_adult)



from abc import ABC

from django.test import TestCase


class GenericTestMixin(TestCase, ABC):  # pragma: no cover
   
    def create_test_imis_instance(self):
        raise NotImplementedError('`test_imis_instance()` must be implemented.')

    def verify_imis_instance(self, imis_obj):
        raise NotImplementedError('`verify_imis_instance()` must be implemented.')

    def create_test_fhir_instance(self):
        raise NotImplementedError('`test_fhir_instance()` must be implemented.')

    def verify_fhir_instance(self, fhir_obj):
        raise NotImplementedError('`verify_fhir_instance()` must be implemented.')



from django.db.models.query import Q
from django.utils.translation import gettext as _
from fhir.resources.R4B.humanname import HumanName

from insuree.models import Insuree, InsureePolicy, Family, FamilyType, ConfirmationType
from policy.models import Policy
from location.models import Location
from api_fhir_r4.configurations import R4IdentifierConfig, GeneralConfiguration
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.converters.locationConverter import LocationConverter

from api_fhir_r4.mapping.groupMapping import GroupTypeMapping, ConfirmationTypeMapping
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.group import Group, GroupMember
from api_fhir_r4.utils import DbManagerUtils
from api_fhir_r4.exceptions import FHIRException


class GroupConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_family, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_family = {}
        # create two obligatory field
        cls.build_fhir_actual(fhir_family, imis_family)
        cls.build_fhir_type(fhir_family, imis_family)
        fhir_family = Group(**fhir_family)
        # then create fhir object as usual
        cls.build_fhir_extentions(fhir_family, imis_family, reference_type)
        cls.build_fhir_identifiers(fhir_family, imis_family)
        cls.build_fhir_pk(fhir_family, imis_family, reference_type=reference_type)
        cls.build_fhir_active(fhir_family, imis_family)
        cls.build_fhir_quantity(fhir_family, imis_family)
        cls.build_fhir_name(fhir_family, imis_family)
        cls.build_fhir_member(fhir_family, imis_family, reference_type)
        return fhir_family

    @classmethod
    def to_imis_obj(cls, fhir_family, audit_user_id):
        errors = []
        fhir_family = Group(**fhir_family)
        imis_family = Family()
        # set uuid to Null so as to use this family object properly in service
        imis_family.uuid = None
        imis_family.audit_user_id = audit_user_id
        cls.build_family_members(imis_family, fhir_family, errors)
        cls.build_imis_extentions(imis_family, fhir_family)
        cls.check_errors(errors)
        return imis_family

    @classmethod
    def get_reference_obj_id(cls, imis_family):
        return imis_family.id

    @classmethod
    def get_fhir_resource_type(cls):
        return Group

    @classmethod
    def build_fhir_pk(cls, fhir_obj, resource, reference_type: str = None):
        if reference_type == ReferenceConverterMixin.CODE_REFERENCE_TYPE:
            fhir_obj.id = resource.head_insuree.chf_id
        else:
            return super().build_fhir_pk(fhir_obj, resource, reference_type)

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        return DbManagerUtils.get_object_or_none(
            Family,
            **cls.get_database_query_id_parameteres_from_reference(reference))


    @classmethod
    def build_human_names(cls, fhir_family, imis_family):
        name = cls.build_fhir_names_for_person(imis_family)
        if type(fhir_family.head) is not list:
            fhir_family.head = [name]
        else:
            fhir_family.head.append(name)

    @classmethod
    def build_family_members(cls, imis_family, fhir_family, errors):
        members = fhir_family.member
        imis_family.members_family = []
        if cls.valid_condition(members is None, _('Missing `member` attribute'), errors) \
                or cls.valid_condition(members is None, _('Missing member should not be empty'), errors):
            return

        for member in members:
            cls._add_member_to_family(imis_family, member.entity, imis_family.members_family)
        
    @classmethod
    def build_fhir_identifiers(cls, fhir_family, imis_family):
        identifiers = []
        cls.build_fhir_uuid_identifier(identifiers, imis_family)
        cls.build_fhir_code_identifier(identifiers, imis_family)
        fhir_family.identifier = identifiers
        cls._validate_fhir_identifier_is_exist(fhir_family)

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_family):
        cls._build_family_head_identifier(identifiers, imis_family)

    @classmethod
    def _build_family_head_identifier(cls, identifiers, imis_family):
        cls._validate_imis_identifier_code(imis_family)
        head_id = cls.build_fhir_identifier(
            imis_family.head_insuree.chf_id,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )
        identifiers.append(head_id)

    @classmethod
    def _add_member_to_family(cls, imis_family, reference, members_family):
        cls._validate_fhir_family_identifier_code(reference)
        insuree = cls._insuree_from_reference(reference)
        if cls._insuree_has_family(insuree):
            raise FHIRException(F"Insuree {insuree} already asigned to family")

        if len(members_family) == 0:
            imis_family.head_insuree = insuree
        members_family.append(imis_family.head_insuree)

    @classmethod
    def build_fhir_name(cls, fhir_family, imis_family):
      if imis_family.head_insuree is not None:
          fhir_family.name = imis_family.head_insuree.last_name
    
    @classmethod
    def build_fhir_actual(cls, fhir_family, imis_family):
        fhir_family['actual'] = True
    
    @classmethod
    def build_fhir_type(cls, fhir_family, imis_family):
        # according to the IMIS profile - always 'Person' value
        fhir_family['type'] = "Person"
        
    @classmethod
    def build_fhir_active(cls, fhir_family, imis_family):
        number_of_active_policy = InsureePolicy.objects.filter(
            Q(insuree__family__uuid=imis_family.uuid),
            Q(policy__status=Policy.STATUS_ACTIVE),
            Q(validity_to__isnull=True)
        ).count()
        fhir_family.active = True if number_of_active_policy > 0 else False

    @classmethod
    def build_fhir_member(cls,fhir_family, imis_family, reference_type):
        fhir_family.member = cls.build_fhir_members(imis_family)

    @classmethod
    def build_fhir_quantity(cls,fhir_family, imis_family):
        quantity = Insuree.objects.filter(family__uuid=imis_family.uuid, validity_to__isnull=True).count()
        fhir_family.quantity = quantity

    @classmethod
    def build_fhir_extentions(cls, fhir_family, imis_family, reference_type):
        fhir_family.extension = []

        def build_extension(fhir_family, imis_family, value):
            extension = Extension.construct()
            if value == "group-address":
                cls._build_extension_address(extension, fhir_family, imis_family, reference_type=reference_type)

            elif value == "group-poverty-status":
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-poverty-status"
                extension.valueBoolean = imis_family.poverty

            elif value == "group-type":
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-type"
                if hasattr(imis_family, "family_type") and imis_family.family_type is not None:
                    GroupTypeMapping.load()
                    display = GroupTypeMapping.group_type[str(imis_family.family_type.code)]
                    system = "CodeSystem/group-type"
                    extension.valueCodeableConcept = cls.build_codeable_concept(code=str(imis_family.family_type.code),
                                                                                system=system)
                    if len(extension.valueCodeableConcept.coding) == 1:
                        extension.valueCodeableConcept.coding[0].display = display
            # group-confirmation
            else:
                cls._build_extension_group_confirmation(extension, fhir_family, imis_family)

            if type(fhir_family.extension) is not list:
                fhir_family.extension = [extension]
            else:
                fhir_family.extension.append(extension)

        # check if family has location
        cls._validate_imis_family_location(imis_family)
        build_extension(fhir_family, imis_family, "group-address")
        if imis_family.poverty is not None:
            build_extension(fhir_family, imis_family, "group-poverty-status")
        if imis_family.family_type is not None:
            build_extension(fhir_family, imis_family, "group-type")
        if imis_family.confirmation_type is not None and imis_family.confirmation_no is not None:
            build_extension(fhir_family, imis_family, "group-confirmation")

    @classmethod
    def build_imis_extentions(cls, imis_family, fhir_family):
        cls._validate_fhir_extension_is_exist(fhir_family)
        for extension in fhir_family.extension:
            if extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-address":
                address = extension.valueAddress
                # insuree use temp address
                if address and address.use == "home" and address.type == "physical":
                    imis_family.address = address.text
                    for ext in address.extension:
                        if "StructureDefinition/address-location-reference" in ext.url:
                                imis_family.location= Location.objects.filter(validity_to__isnull=True,**LocationConverter.get_database_query_id_parameteres_from_reference(ext.valueReference)).first()

            elif extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-poverty-status":
                imis_family.poverty = extension.valueBoolean

            elif extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-type":
                try:
                    imis_family.family_type = FamilyType.objects.get(code=extension.valueCodeableConcept.coding[0].code)
                except:
                    imis_family.family_type = None

            elif extension.url == f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-confirmation":
                try:
                    for ext in extension.extension:
                        if ext.url == "number":
                            fhir_family.confirmation_no = ext.valueString
                        if ext.url == "type":
                            fhir_family.confirmation_type = ConfirmationType.objects.get(
                                code=ext.valueCodeableConcept.coding[0].code)
                except:
                    imis_family.confirmation_no = None
                    imis_family.confirmation_type = None
            else:
                pass
        cls._validate_imis_family_location(imis_family)



    @classmethod
    def _build_extension_address(cls, extension, fhir_family, imis_family, reference_type):
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-address"
        family_address = cls.build_fhir_address(imis_family.address, "home", "physical")
        if imis_family.location:
            family_address.state = imis_family.location.parent.parent.parent.name
            family_address.district = imis_family.location.parent.parent.name
            # municipality extension
            extension_address = Extension.construct()
            extension_address.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
            extension_address.valueString = imis_family.location.parent.name
            family_address.extension = [extension_address]

            # address location reference extension
            extension_address = Extension.construct()
            extension_address.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-location-reference"
            extension_address.valueReference = LocationConverter\
                .build_fhir_resource_reference(imis_family.location, 'Location', reference_type=reference_type)
            family_address.extension.append(extension_address)
            family_address.city = imis_family.location.name
        extension.valueAddress = family_address

    @classmethod
    def _build_extension_group_confirmation(cls, extension, fhir_family, imis_family):
        nested_extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-confirmation"
        if hasattr(imis_family, "confirmation_type") and imis_family.confirmation_type:
            if hasattr(imis_family, "confirmation_no") and imis_family.confirmation_no:
                # add number extension
                nested_extension.url = "number"
                nested_extension.valueString = imis_family.confirmation_no
                extension.extension = [nested_extension]
                # add identifier extension
                nested_extension = Extension.construct()
                nested_extension.url = "type"
                system = "CodeSystem/group-confirmation-type" 
                ConfirmationTypeMapping.load()
                display = ConfirmationTypeMapping.confirmation_type[str(imis_family.confirmation_type.code)]
                nested_extension.valueCodeableConcept = cls.build_codeable_concept(
                    code=imis_family.confirmation_type.code, system=system)
                if len(nested_extension.valueCodeableConcept.coding) == 1:
                    nested_extension.valueCodeableConcept.coding[0].display = display
                extension.extension.append(nested_extension)

    @classmethod
    def build_fhir_names_for_person(cls, person_obj):
        try:
            return HumanName(**{
                'use': 'usual',
                'family': person_obj.last_name,
                'given': [person_obj.other_names]
            })
        except AttributeError as e:
            raise FHIRException([_('Missing `last_name` and `other_names` for IMIS object')]) from e

    @classmethod
    def build_fhir_members(cls, family):
        family_insurees = family.members.all()
        members = [cls._create_group_member(member) for member in family_insurees]
        return members

    @classmethod
    def get_reference_obj_uuid(cls, imis_family: Family):
        return imis_family.uuid

    @classmethod
    def get_reference_obj_code(cls, imis_family: Family):
        return imis_family.head_insuree.chf_id

    # fhir validations
    @classmethod
    def _validate_fhir_identifier_is_exist(cls, fhir_family):
        if not fhir_family.identifier or len(fhir_family.identifier) == 0:
            raise FHIRException(
                _('FHIR Group entity without identifier')
            )

    # fhir validations
    @classmethod
    def _validate_fhir_extension_is_exist(cls, fhir_family):
        if not fhir_family.extension or len(fhir_family.extension) == 0:
            raise FHIRException(
                _('At least one extension with address is required')
            )

    @classmethod
    def _validate_fhir_family_identifier_code(cls, fhir_family_identifier_code):
        if not fhir_family_identifier_code:
            raise FHIRException(
                _('Family Group FHIR without code - this field is obligatory')
            )

    # imis validations
    @classmethod
    def _validate_imis_identifier_code(cls, imis_family):
        if not imis_family.head_insuree.chf_id:
            raise FHIRException(
                _('Family %(family_uuid)s without code') % {'family_uuid': imis_family.uuid}
            )

    @classmethod
    def _validate_imis_family_location(cls, imis_family):
        if not imis_family.location:
            if imis_family.uuid:
                raise FHIRException(
                    _('Family %(family_uuid)s without location') % {'family_uuid': imis_family.uuid}
                )
            else:
                raise FHIRException(
                    _('new Family without location')
                )

    @classmethod
    def _create_group_member(cls, insuree):
        # Due to circular dependency import has to be done inside of method
        from api_fhir_r4.converters import PatientConverter
        reference = PatientConverter.build_fhir_resource_reference(
            insuree,
            type='Patient',
            display=insuree.chf_id
        )
        return GroupMember(entity=reference)

    @classmethod
    def _insuree_from_reference(cls, reference):
        from api_fhir_r4.converters import PatientConverter
        insuree = PatientConverter.get_imis_obj_by_fhir_reference(reference)
        if insuree is None:
            raise FHIRException(f"Invalid Family member reference. "
                                f"Insuree with identifier `{reference}` does not exist")
        return insuree

    @classmethod
    def _insuree_has_family(cls, insuree):
        return insuree.family is not None



from uuid import UUID

from insuree.models import Gender, Insuree, Profession
from location.models import Location
from location.test_helpers import create_test_location,create_test_village
from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig, R4MaritalConfig
from api_fhir_r4.converters import PatientConverter
from api_fhir_r4.mapping.patientMapping import PatientProfessionMapping
from fhir.resources.R4B.address import Address
from fhir.resources.R4B.attachment import Attachment
from fhir.resources.R4B.contactpoint import ContactPoint
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.humanname import HumanName
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.patient import Patient
from fhir.resources.R4B.reference import Reference
from api_fhir_r4.models.imisModelEnums import ContactPointSystem, ContactPointUse
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils

from insuree.test_helpers import create_test_insuree


class PatientTestMixin(GenericTestMixin):
    _TEST_LAST_NAME = "TEST_LAST_NAME"
    _TEST_OTHER_NAME = "TEST_OTHER_NAME"
    _TEST_IS_HEAD = False
    _TEST_INSUREE_DOB = "1990-03-24"
    _TEST_INSUREE_ID = 10000
    _TEST_INSUREE_UUID = "0a60f36c-62eb-11ea-bb93-93ec0339a3dd"
    _TEST_INSUREE_CHFID = "827192671"
    _TEST_PASSPORT = "TEST_PASSPORT"
    _TEST_GENDER_CODE = "M"
    _TEST_GENDER = None
    _TEST_PROFESSION = None
    _TEST_CARD_ISSUED = False
    _TEST_PHONE = "813-996-476"
    _TEST_EMAIL = "TEST@TEST.com"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_INSUREE_MOCKED_UUID = "7240daef-5f8f-4b0f-9042-b221e66f184a"
    _TEST_GROUP_UUID = "8e33033a-9f60-43ad-be3e-3bfeb992aae5"
    _TEST_VILLAGE_CODE = "RTDTMTVT"
    _TEST_VILLAGE_UUID = "69a55f2d-ee34-4193-be0e-2b6a361797bd"
    _TEST_VILLAGE_NAME = "TEST_NAME"
    _TEST_PHOTO_FOLDER = "PhotoTest"
    _TEST_MOCKED_PHOTO = "TESTTEST"
    _TEST_MOCKED_PHOTO_TYPE = "png"
    _TEST_MOCKED_PHOTO_CREATION = "2021-03-27"
    _TEST_PHOTO_TITLE = "photo_test"
    test_insuree= None
    test_village= None
    test_region= None
    test_ward = None
    test_district = None
    sub_str = {}
    @classmethod
    def setUpTestData(cls):
        cls._TEST_GENDER = Gender()
        cls._TEST_GENDER.code = cls._TEST_GENDER_CODE
        cls._TEST_PROFESSION = Profession.objects.get(id=4)

        if cls.test_region is None:
            
            cls.test_village  =create_test_village( custom_props={"code":cls._TEST_VILLAGE_CODE,"name":cls._TEST_VILLAGE_NAME})
            cls.test_ward =cls.test_village.parent
            cls.test_region =cls.test_village.parent.parent.parent
            cls.test_district = cls.test_village.parent.parent
            
        cls.test_insuree = create_test_insuree(
            custom_props = {
                "last_name":cls._TEST_LAST_NAME,
                "other_names":cls._TEST_OTHER_NAME,
            "id" : cls._TEST_INSUREE_ID,
            "uuid" : cls._TEST_INSUREE_UUID,
            "chf_id" : cls._TEST_INSUREE_CHFID,
            "passport" : cls._TEST_PASSPORT,
            "dob" : TimeUtils.str_to_date(cls._TEST_INSUREE_DOB),
            "gender" : cls._TEST_GENDER,
            "marital": "D",
            "phone" : cls._TEST_PHONE,
            "email" : cls._TEST_EMAIL,
            "current_address" : cls._TEST_ADDRESS,
            "current_village" : cls.test_village,
            "profession" : cls._TEST_PROFESSION,
            "card_issued" : cls._TEST_CARD_ISSUED,
            }
        )
        cls.sub_str[cls._TEST_INSUREE_CHFID] = cls.test_insuree.chf_id
        cls.sub_str[cls._TEST_INSUREE_UUID] = cls.test_insuree.uuid
        cls.sub_str[cls._TEST_VILLAGE_UUID] = cls.test_village.uuid
        cls.sub_str[cls._TEST_GROUP_UUID] = cls.test_insuree.family.uuid
        

        
            
    def create_test_imis_instance(self):
        return self.test_insuree

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_LAST_NAME, imis_obj.last_name)
        self.assertEqual(self._TEST_OTHER_NAME, imis_obj.other_names)
        self.assertEqual(self._TEST_INSUREE_CHFID, imis_obj.chf_id)
        expected_date = TimeUtils.str_to_date(self._TEST_INSUREE_DOB)
        self.assertEqual(expected_date, imis_obj.dob)
        self.assertEqual("D", imis_obj.marital)
        self.assertEqual(self._TEST_PHONE, imis_obj.phone)
        self.assertEqual(self._TEST_EMAIL, imis_obj.email)
        self.assertEqual(self._TEST_ADDRESS, imis_obj.current_address)
        self.assertEqual(self._TEST_IS_HEAD, imis_obj.head)
        self.assertEqual(self._TEST_PROFESSION.profession, imis_obj.profession.profession)
        self.assertEqual(self._TEST_CARD_ISSUED, imis_obj.card_issued)

    def create_test_fhir_instance(self):

        fhir_patient = Patient.construct()
        name = HumanName.construct()
        name.family = self.test_insuree.last_name
        name.given = [ self.test_insuree.other_names]
        name.use = "usual"
        fhir_patient.name = [name]
        identifiers = []
        chf_id = PatientConverter.build_fhir_identifier(
            self.test_insuree.chf_id,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )

        identifiers.append(chf_id)

        fhir_patient.identifier = identifiers
        fhir_patient.birthDate =  self.test_insuree.dob
        fhir_patient.gender = "male"
        fhir_patient.maritalStatus = PatientConverter.build_codeable_concept(
            R4MaritalConfig.get_fhir_divorced_code(),
            R4MaritalConfig.get_fhir_marital_status_system())
        telecom = []
        phone = PatientConverter.build_fhir_contact_point(
            self._TEST_PHONE,
            ContactPointSystem.PHONE,
            ContactPointUse.HOME
        )
        telecom.append(phone)
        email = PatientConverter.build_fhir_contact_point(
            self._TEST_EMAIL,
            ContactPointSystem.EMAIL,
            ContactPointUse.HOME
        )
        telecom.append(email)
        fhir_patient.telecom = telecom

        # family slice - required
        family_address = PatientConverter.build_fhir_address(self._TEST_ADDRESS, "home", "physical")
        family_address.state = self.test_village.parent.parent.parent.name
        family_address.district = self.test_village.parent.parent.name

        # municipality extension
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
        extension.valueString = self.test_village.parent.name
        family_address.extension = [extension]

        # address location reference extension
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-location-reference"
        reference_location = Reference.construct()
        reference_location.reference = F"Location/{self.test_village.uuid}"
        extension.valueReference = reference_location
        family_address.extension.append(extension)

        family_address.city = self.test_village.name

        addresses = [family_address]

        # Commented out, as single address extension is required
        # # insuree slice
        # current_address = PatientConverter.build_fhir_address(self._TEST_ADDRESS, "temp", "physical")
        # current_address.state = self.test_village.parent.parent.parent.name
        # current_address.district = self.test_village.parent.parent.name
        #
        # # municipality extension
        # extension = Extension.construct()
        # extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
        # extension.valueString = self.test_village.parent.name
        # current_address.extension = [extension]
        #
        # # address location reference extension
        # extension = Extension.construct()
        # extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-location-reference"
        # reference_location = Reference.construct()
        # reference_location.reference = F"Location/{self.test_village.uuid}"
        # extension.valueReference = reference_location
        # current_address.extension.append(extension)
        # current_address.city = self.test_village.name
        #
        # addresses.append(current_address)
        fhir_patient.address = addresses

        # test mocked_photo
        photo = Attachment.construct()
        photo.data = self._TEST_MOCKED_PHOTO
        photo.creation = self._TEST_MOCKED_PHOTO_CREATION
        photo.contentType = self._TEST_MOCKED_PHOTO_TYPE
        photo.title = self._TEST_PHOTO_TITLE
        if type(fhir_patient.photo) is not list:
            fhir_patient.photo = [photo]
        else:
            fhir_patient.photo.append(photo)

        # extensions
        fhir_patient.extension = []

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-is-head"
        extension.valueBoolean = self._TEST_IS_HEAD
        fhir_patient.extension = [extension]

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-card-issued"
        extension.valueBoolean = self._TEST_CARD_ISSUED
        fhir_patient.extension.append(extension)

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-profession"
        PatientProfessionMapping.load()
        display = PatientProfessionMapping.patient_profession[str(self._TEST_PROFESSION.id)]
        system = "CodeSystem/patient-profession"
        extension.valueCodeableConcept = PatientConverter.build_codeable_concept(code=str(self._TEST_PROFESSION.id),
                                                                                 system=system)
        if len(extension.valueCodeableConcept.coding) == 1:
            extension.valueCodeableConcept.coding[0].display = display
        fhir_patient.extension.append(extension)

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-group-reference"
        reference_group = Reference.construct()
        reference_group.reference = F"Group/{self.test_insuree.family.uuid}"
        extension.valueReference = reference_group
        fhir_patient.extension.append(extension)

        return fhir_patient

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(1, len(fhir_obj.name))
        human_name = fhir_obj.name[0]
        self.assertTrue(isinstance(human_name, HumanName))
        self.assertEqual(self._TEST_OTHER_NAME, human_name.given[0])
        self.assertEqual(self._TEST_LAST_NAME, human_name.family)
        self.assertEqual("usual", human_name.use)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = PatientConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(str(self._TEST_INSUREE_CHFID), identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code() and not isinstance(identifier.value, UUID):
                self.assertEqual(str(self._TEST_INSUREE_UUID), identifier.value)
        self.assertEqual(self._TEST_INSUREE_DOB, fhir_obj.birthDate.isoformat())
        self.assertEqual("male", fhir_obj.gender)
        marital_code = PatientConverter.get_first_coding_from_codeable_concept(fhir_obj.maritalStatus).code
        self.assertEqual(R4MaritalConfig.get_fhir_divorced_code(), marital_code)
        self.assertEqual(2, len(fhir_obj.telecom))
        for telecom in fhir_obj.telecom:
            self.assertTrue(isinstance(telecom, ContactPoint))
            if telecom.system == "phone":
                self.assertEqual(self._TEST_PHONE, telecom.value)
            elif telecom.system == "email":
                self.assertEqual(self._TEST_EMAIL, telecom.value)
        for address in fhir_obj.address:
            self.assertTrue(isinstance(address, Address))
            if address.use == "home":
                no_of_extensions = len(address.extension)
                self.assertEqual(2, no_of_extensions)
                self.assertEqual("home", address.use)
                self.assertEqual(self._TEST_VILLAGE_NAME, address.city)
            elif address.use == "temp":
                self.assertEqual("temp", address.use)
                self.assertEqual(self._TEST_VILLAGE_NAME, address.city)
        for extension in fhir_obj.extension:
            self.assertTrue(isinstance(extension, Extension))
            if "patient-group-reference" in extension.url:
                self.assertIn(str(self.test_insuree.family.uuid), extension.valueReference.reference)
            if "patient-card-issue" in extension.url:
                self.assertEqual(self._TEST_CARD_ISSUED, extension.valueBoolean)
            if "patient-is-head" in extension.url:
                self.assertEqual(self._TEST_IS_HEAD, extension.valueBoolean)
            if "patient-profession" in extension.url:
                self.assertEqual(self._TEST_PROFESSION.profession, extension.valueCodeableConcept.coding[0].display)



ERROR_BAD_STATUS = "Bad value in paymentStatus, should be either 'active', 'cancelled', draft', 'entered-in-error'"
ERROR_BAD_PAYMENT_STATUS = "Bad value in paymentStatus, should be either 'paid' or 'cleared'"
ERROR_BAD_TYPE_REQUEST = "Bad value in type in request, should be either 'bill' or 'invoice'"



from django.utils.translation import gettext

from api_fhir_r4.converters import BaseFHIRConverter
from api_fhir_r4.exceptions import FHIRRequestProcessException
from fhir.resources.R4B.humanname import HumanName

from api_fhir_r4.models.imisModelEnums import ContactPointSystem, ContactPointUse


class PersonConverterMixin(object):

    @classmethod
    def build_fhir_names_for_person(cls, person_obj):
        if not hasattr(person_obj, 'last_name') and not hasattr(person_obj, 'other_names'):
            raise FHIRRequestProcessException([gettext('Missing `last_name` and `other_names` for IMIS object')])
        name = HumanName.construct()
        name.use = "usual"
        name.family = person_obj.last_name
        name.given = [person_obj.other_names]
        return name

    @classmethod
    def build_imis_last_and_other_name(cls, names):
        last_name = None
        other_names = None
        if isinstance(names, list):
            for name in names:
                if name.use == "usual":
                    last_name = name.family
                    given_names = name.given
                    if given_names and len(given_names) > 0:
                        other_names = given_names[0]
                    break
        return last_name, other_names

    @classmethod
    def build_fhir_telecom_for_person(cls, phone=None, email=None):
        telecom = []
        if phone:
            phone = BaseFHIRConverter.build_fhir_contact_point(phone, ContactPointSystem.PHONE, ContactPointUse.HOME)
            telecom.append(phone)
        if email:
            email = BaseFHIRConverter.build_fhir_contact_point(email, ContactPointSystem.EMAIL, ContactPointUse.HOME)
            telecom.append(email)
        return telecom

    @classmethod
    def build_imis_phone_num_and_email(cls, telecom):
        phone = None
        email = None
        if telecom is not None:
            for contact_point in telecom:
                if contact_point.system == ContactPointSystem.PHONE.value:
                    phone = contact_point.value
                elif contact_point.system == ContactPointSystem.EMAIL.value:
                    email = contact_point.value
        return phone, email



import sys

from api_fhir_r4.defaultConfig import DEFAULT_CFG


class BaseConfiguration(object):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_config(cls):
        module_name = "api_fhir_r4"
        return sys.modules[module_name]

    @classmethod
    def get_config_attribute(cls, attribute, cfg=None):
        conf = cls.get_config()
        if conf and hasattr(conf, attribute):
            return conf.__getattribute__(attribute)
        else:
            from core.models import ModuleConfiguration
            cfg = cfg or ModuleConfiguration.get_or_default("api_fhir_r4", DEFAULT_CFG)
            cls.build_configuration(cfg)
            return conf.__getattribute__(attribute)


class IdentifierConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_identifier_type_system(cls):
        raise NotImplementedError('`get_fhir_identifier_type_system()` must be implemented.')

    @classmethod
    def get_fhir_uuid_type_code(cls):
        raise NotImplementedError('`get_fhir_id_type_code()` must be implemented.')

    @classmethod
    def get_fhir_id_type_code(cls):
        raise NotImplementedError('`get_fhir_id_type_code()` must be implemented.')

    @classmethod
    def get_fhir_chfid_type_code(cls):
        raise NotImplementedError('`get_fhir_chfid_type_code()` must be implemented.')

    @classmethod
    def get_fhir_passport_type_code(cls):
        raise NotImplementedError('`get_fhir_passport_type_code()` must be implemented.')

    @classmethod
    def get_fhir_facility_id_type(cls):
        raise NotImplementedError('`get_fhir_facility_id_type()` must be implemented.')

    @classmethod
    def get_fhir_claim_admin_code_type(cls):
        raise NotImplementedError('`get_fhir_claim_admin_code_type()` must be implemented.')

    @classmethod
    def get_fhir_claim_code_type(cls):
        raise NotImplementedError('`get_fhir_claim_code_type()` must be implemented.')


class LocationConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_location_site_type_system(cls):
        raise NotImplementedError('`get_fhir_location_site_type_system()` must be implemented.')

    @classmethod
    def get_fhir_code_for_hospital(cls):
        raise NotImplementedError('`get_fhir_code_for_hospital()` must be implemented.')

    @classmethod
    def get_fhir_code_for_dispensary(cls):
        raise NotImplementedError('`get_fhir_code_for_dispensary()` must be implemented.')

    @classmethod
    def get_fhir_code_for_health_center(cls):
        raise NotImplementedError('`get_fhir_code_for_health_center()` must be implemented.')

    @classmethod
    def get_fhir_location_physical_type_system(cls):
        raise NotImplementedError('`get_fhir_location_physical_type_system()` must be implemented.')

    @classmethod
    def get_fhir_code_for_region(cls):
        raise NotImplementedError('`get_fhir_code_for_region()` must be implemented.')

    @classmethod
    def get_fhir_code_for_district(cls):
        raise NotImplementedError('`get_fhir_code_for_district()` must be implemented.')

    @classmethod
    def get_fhir_code_for_ward(cls):
        raise NotImplementedError('`get_fhir_code_for_ward()` must be implemented.')

    @classmethod
    def get_fhir_code_for_village(cls):
        raise NotImplementedError('`get_fhir_code_for_village()` must be implemented.')

    @classmethod
    def get_fhir_hf_service_type_system(cls):
        raise NotImplementedError('`get_fhir_hf_service_type_system()` must be implemented')

    @classmethod
    def get_fhir_code_for_in_patient(cls):
        raise NotImplementedError('`get_fhir_code_for_in_patient()` must be implemented')

    @classmethod
    def get_fhir_code_for_out_patient(cls):
        raise NotImplementedError('`get_fhir_code_for_out_patient()` must be implemented')

    @classmethod
    def get_fhir_code_for_both(cls):
        raise NotImplementedError('`get_fhir_code_for_both()` must be implemented')

    @classmethod
    def get_fhir_code_for_active(cls):
        raise NotImplementedError('`get_fhir_code_for_active()` must be implemented')

    @classmethod
    def get_subscribe_location_signal(cls):
        raise NotImplementedError('`get_subscribe_location_signal()` must be implemented')


class MaritalConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_marital_status_system(cls):
        raise NotImplementedError('`get_fhir_marital_status_system()` must be implemented.')

    @classmethod
    def get_fhir_married_code(cls):
        raise NotImplementedError('`get_fhir_married_code()` must be implemented.')

    @classmethod
    def get_fhir_never_married_code(cls):
        raise NotImplementedError('`get_fhir_never_married_code()` must be implemented.')

    @classmethod
    def get_fhir_divorced_code(cls):
        raise NotImplementedError('`get_fhir_divorced_code()` must be implemented.')

    @classmethod
    def get_fhir_widowed_code(cls):
        raise NotImplementedError('`get_fhir_widowed_code()` must be implemented.')

    @classmethod
    def get_fhir_unknown_marital_status_code(cls):
        raise NotImplementedError('`get_fhir_unknown_marital_status_code()` must be implemented.')


class IssueTypeConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_code_for_exception(cls):
        raise NotImplementedError('`get_fhir_code_for_exception()` must be implemented.')

    @classmethod
    def get_fhir_code_for_not_found(cls):
        raise NotImplementedError('`get_fhir_code_for_not_found()` must be implemented.')

    @classmethod
    def get_fhir_code_for_informational(cls):
        raise NotImplementedError('`get_fhir_code_for_informational()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_general_adjudication_code(cls):
        raise NotImplementedError('`get_fhir_claim_item_general_adjudication_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_rejected_reason_adjudication_code(cls):
        raise NotImplementedError('`get_fhir_claim_item_rejected_reason_adjudication_code()` must be implemented.')


class ClaimConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_claim_information_guarantee_id_code(cls):
        raise NotImplementedError('`get_fhir_claim_information_guarantee_id_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_information_explanation_code(cls):
        raise NotImplementedError('`get_fhir_claim_information_explanation_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_explanation_code(cls):
        raise NotImplementedError('`get_fhir_claim_item_explanation_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_code(cls):
        raise NotImplementedError('`get_fhir_claim_item_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_service_code(cls):
        raise NotImplementedError('`get_fhir_claim_service_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_status_rejected_code(cls):
        raise NotImplementedError('`get_fhir_claim_status_rejected_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_status_entered_code(cls):
        raise NotImplementedError('`get_fhir_claim_status_entered_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_status_checked_code(cls):
        raise NotImplementedError('`get_fhir_claim_status_checked_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_status_processed_code(cls):
        raise NotImplementedError('`get_fhir_claim_status_processed_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_status_valuated_code(cls):
        raise NotImplementedError('`get_fhir_claim_status_valuated_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_status_passed_code(cls):
        raise NotImplementedError('`get_fhir_claim_item_status_passed_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_status_rejected_code(cls):
        raise NotImplementedError('`get_fhir_claim_item_status_rejected_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_general_adjudication_code(cls):
        raise NotImplementedError('`get_fhir_claim_item_general_adjudication_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_rejected_reason_adjudication_code(cls):
        raise NotImplementedError('`get_fhir_claim_item_rejected_reason_adjudication_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_organization_code(cls):
        raise NotImplementedError('`get_fhir_claim_organization_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_attachment_code(cls):
        raise NotImplementedError('`get_fhir_claim_attachment_code()` must be implemented.')

    @classmethod
    def get_fhir_claim_visit_type_system(cls):
        raise NotImplementedError('`get_fhir_claim_visit_type_system()` must be implemented.')

    @classmethod
    def get_fhir_claim_supporting_info_system(cls):
        raise NotImplementedError('`get_fhir_claim_supporting_info_system()` must be implemented.')

    @classmethod
    def get_fhir_item_reference_extension_system(cls):
        raise NotImplementedError('`get_fhir_item_reference_extension_system()` must be implemented.')

    @classmethod
    def get_fhir_claim_item_category_system(cls):
        raise NotImplementedError('`get_fhir_claim_item_category_system()` must be implemented.')

    @classmethod
    def get_fhir_claim_diagnosis_system(cls):
        raise NotImplementedError('`get_fhir_claim_diagnosis_system()` must be implemented.')

    @classmethod
    def get_fhir_claim_attachment_system(cls):
        raise NotImplementedError('`get_fhir_claim_attachment_system()` must be implemented.')

    @classmethod
    def get_allowed_fhir_claim_attachment_mime_types_regex(cls):
        raise NotImplementedError('`get_allowed_fhir_claim_attachment_mime_types_regex()` must be implemented.')


class CoverageEligibilityConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_financial_code(cls):
        raise NotImplementedError('`get_fhir_financial_code()` must be implemented.')

    @classmethod
    def get_fhir_item_code(cls):
        raise NotImplementedError('`get_fhir_item_code()` must be implemented.')

    @classmethod
    def get_fhir_service_code(cls):
        raise NotImplementedError('`get_fhir_service_code()` must be implemented.')

    @classmethod
    def get_fhir_total_admissions_code(cls):
        raise NotImplementedError('`get_fhir_total_admissions_code()` must be implemented.')

    @classmethod
    def get_fhir_total_visits_code(cls):
        raise NotImplementedError('`get_fhir_total_visits_code()` must be implemented.')

    @classmethod
    def get_fhir_total_consultations_code(cls):
        raise NotImplementedError('`get_fhir_total_consultations_code()` must be implemented.')

    @classmethod
    def get_fhir_total_surgeries_code(cls):
        raise NotImplementedError('`get_fhir_total_surgeries_code()` must be implemented.')

    @classmethod
    def get_fhir_total_deliveries_code(cls):
        raise NotImplementedError('`get_fhir_total_deliveries_code()` must be implemented.')

    @classmethod
    def get_fhir_total_antenatal_code(cls):
        raise NotImplementedError('`get_fhir_total_antenatal_code()` must be implemented.')

    @classmethod
    def get_fhir_consultation_amount_code(cls):
        raise NotImplementedError('`get_fhir_consultation_amount_code()` must be implemented.')

    @classmethod
    def get_fhir_surgery_amount_code(cls):
        raise NotImplementedError('`get_fhir_surgery_amount_code()` must be implemented.')

    @classmethod
    def get_fhir_delivery_amount_code(cls):
        raise NotImplementedError('`get_fhir_delivery_amount_code()` must be implemented.')

    @classmethod
    def get_fhir_hospitalization_amount_code(cls):
        raise NotImplementedError('`get_fhir_hospitalization_amount_code()` must be implemented.')

    @classmethod
    def get_fhir_antenatal_amount_code(cls):
        raise NotImplementedError('`get_fhir_antenatal_amount_code()` must be implemented.')

    @classmethod
    def get_fhir_service_left_code(cls):
        raise NotImplementedError('`get_fhir_service_left_code()` must be implemented.')

    @classmethod
    def get_fhir_item_left_code(cls):
        raise NotImplementedError('`get_fhir_item_left_code()` must be implemented.')

    @classmethod
    def get_fhir_is_item_ok_code(cls):
        raise NotImplementedError('`get_fhir_is_item_ok_code()` must be implemented.')

    @classmethod
    def get_fhir_is_service_ok_code(cls):
        raise NotImplementedError('`get_fhir_is_service_ok_code()` must be implemented.')

    @classmethod
    def get_fhir_balance_code(cls):
        raise NotImplementedError('`get_fhir_balance_code()` must be implemented.')

    @classmethod
    def get_fhir_balance_default_category(cls):
        raise NotImplementedError('`get_fhir_balance_default_category()` must be implemented.')

    @classmethod
    def get_fhir_status_map(cls):
        raise NotImplementedError('`get_fhir_status_map()` must be implemented.')


class CommunicationRequestConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_care_rendered_code(cls):
        raise NotImplementedError('`get_fhir_care_rendered_code()` must be implemented.')

    @classmethod
    def get_fhir_payment_asked_code(cls):
        raise NotImplementedError('`get_fhir_payment_asked_code()` must be implemented.')

    @classmethod
    def get_fhir_drug_prescribed_code(cls):
        raise NotImplementedError('`get_fhir_drug_prescribed_code()` must be implemented.')

    @classmethod
    def get_fhir_drug_received_code(cls):
        raise NotImplementedError('`get_fhir_care_rendered_code()` must be implemented.')

    @classmethod
    def get_fhir_asessment_code(cls):
        raise NotImplementedError('`get_fhir_asessment_code()` must be implemented.')


class CoverageConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_family_reference_code(cls):
        raise NotImplementedError('`get_family_reference_code()` must be implemented.')

    @classmethod
    def get_status_idle_code(cls):
        raise NotImplementedError('`get_status_idle_code()` must be implemented.')

    @classmethod
    def get_status_active_code(cls):
        raise NotImplementedError('`get_status_active_code()` must be implemented.')

    @classmethod
    def get_status_suspended_code(cls):
        raise NotImplementedError('`get_status_suspended_code()` must be implemented.')

    @classmethod
    def get_status_expired_code(cls):
        raise NotImplementedError('`get_status_expired_code()` must be implemented.')

    @classmethod
    def get_item_code(cls):
        raise NotImplementedError('`get_item_code()` must be implemented.')

    @classmethod
    def get_service_code(cls):
        raise NotImplementedError('`get_service_code()` must be implemented.')

    @classmethod
    def get_practitioner_role_code(cls):
        raise NotImplementedError('`get_practitioner_role_code()` must be implemented.')

    @classmethod
    def get_product_code(cls):
        raise NotImplementedError('`get_product_code()` must be implemented.')

    @classmethod
    def get_enroll_date_code(cls):
        raise NotImplementedError('`get_enroll_date_code()` must be implemented.')

    @classmethod
    def get_effective_date_code(cls):
        raise NotImplementedError('`get_effective_date_code()` must be implemented.')


class InvoiceConfiguration(BaseConfiguration):
    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_invoice_type_system(cls):
        raise NotImplementedError('`get_fhir_invoice_type_system()` must be implemented.')

    @classmethod
    def get_fhir_invoice_charge_item_system(cls):
        raise NotImplementedError('`get_fhir_invoice_charge_item_system()` must be implemented.')

    @classmethod
    def get_fhir_bill_type_system(cls):
        raise NotImplementedError('`get_fhir_invoice_type_system()` must be implemented.')

    @classmethod
    def get_fhir_bill_charge_item_system(cls):
        raise NotImplementedError('`get_fhir_invoice_charge_item_system()` must be implemented.')

    @classmethod
    def get_subscribe_invoice_signal(cls):
        raise NotImplementedError('`get_subscribe_invoice_signal()` must be implemented')


class OrganisationConfiguration(BaseConfiguration):
    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_ph_organisation_type(cls):
        raise NotImplementedError('`get_fhir_ph_organisation_type()` must be implemented.')

    @classmethod
    def get_fhir_address_municipality_extension_system(cls):
        raise NotImplementedError('`get_fhir_address_municipality_extension_system()` must be implemented.')

    @classmethod
    def get_fhir_location_reference_extension_system(cls):
        raise NotImplementedError('`get_fhir_location_reference_extension_system()` must be implemented.')

    @classmethod
    def get_fhir_ph_organisation_legal_form_extension_system(cls):
        raise NotImplementedError('`get_fhir_ph_organisation_legal_form_extension_system()` must be implemented.')

    @classmethod
    def get_fhir_ph_organisation_activity_extension_system(cls):
        raise NotImplementedError('`get_fhir_ph_organisation_activity_extension_system()` must be implemented.')

    @classmethod
    def get_fhir_ph_organisation_legal_form_code_system(cls):
        raise NotImplementedError('`get_fhir_ph_organisation_legal_form_code_system()` must be implemented.')

    @classmethod
    def get_fhir_ph_organisation_activity_code_system(cls):
        raise NotImplementedError('`get_fhir_ph_organisation_activity_code_system()` must be implemented.')

    @classmethod
    def get_fhir_ph_organisation_type_system(cls):
        raise NotImplementedError('`get_fhir_ph_organisation_type_system()` must be implemented.')


class SubscriptionConfiguration(BaseConfiguration):
    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_sub_search_perms(cls):
        raise NotImplementedError('`get_fhir_sub_search_perms()` must be implemented.')

    @classmethod
    def get_fhir_sub_create_perms(cls):
        raise NotImplementedError('`get_fhir_sub_create_perms()` must be implemented.')

    @classmethod
    def get_fhir_sub_update_perms(cls):
        raise NotImplementedError('`get_fhir_sub_update_perms()` must be implemented.')

    @classmethod
    def get_fhir_sub_delete_perms(cls):
        raise NotImplementedError('`get_fhir_sub_delete_perms()` must be implemented.')

    @classmethod
    def get_fhir_subscription_channel_rest_hook(cls):
        raise NotImplementedError('`get_fhir_subscription_channel_rest_hook()` must be implemented.')

    @classmethod
    def get_fhir_subscription_status_off(cls):
        raise NotImplementedError('`get_fhir_subscription_status_off()` must be implemented.')

    @classmethod
    def get_fhir_subscription_status_active(cls):
        raise NotImplementedError('`get_fhir_subscription_status_active()` must be implemented.')

    @classmethod
    def get_fhir_sub_criteria_key_resource(cls):
        raise NotImplementedError('`get_fhir_sub_criteria_key_resource()` must be implemented.')

    @classmethod
    def get_fhir_sub_criteria_key_resource_type(cls):
        raise NotImplementedError('`get_fhir_sub_criteria_key_resource_type()` must be implemented.')


class PaymentNoticeConfiguration(BaseConfiguration):
    @classmethod
    def build_configuration(cls, cfg):
        raise NotImplementedError('`build_configuration()` must be implemented.')

    @classmethod
    def get_fhir_payment_notice_status_active(cls):
        raise NotImplementedError('`get_fhir_payment_notice_status_active()` must be implemented.')

    @classmethod
    def get_fhir_payment_notice_status_cancelled(cls):
        raise NotImplementedError('`get_fhir_payment_notice_status_cancelled()` must be implemented.')

    @classmethod
    def get_fhir_payment_notice_status_draft(cls):
        raise NotImplementedError('`get_fhir_payment_notice_status_draft()` must be implemented.')

    @classmethod
    def get_fhir_payment_notice_status_entered_in_error(cls):
        raise NotImplementedError('`get_fhir_payment_notice_status_entered_in_error()` must be implemented.')

    @classmethod
    def get_fhir_payment_notice_payment_status_paid(cls):
        raise NotImplementedError('`get_fhir_payment_notice_payment_status_paid()` must be implemented.')

    @classmethod
    def get_fhir_payment_notice_payment_status_cleared(cls):
        raise NotImplementedError('`get_fhir_payment_notice_payment_status_cleared()` must be implemented.')


class BaseApiFhirConfiguration(BaseConfiguration):  # pragma: no cover

    @classmethod
    def build_configuration(cls, cfg):
        cls.get_identifier_configuration().build_configuration(cfg)
        cls.get_location_type_configuration().build_configuration(cfg)
        cls.get_marital_type_configuration().build_configuration(cfg)
        cls.get_issue_type_configuration().build_configuration(cfg)
        cls.get_claim_configuration().build_configuration(cfg)
        cls.get_coverage_eligibility_configuration().build_configuration(cfg)
        cls.get_communication_request_configuration().build_configuration(cfg)
        cls.get_coverage_configuration().build_configuration(cfg)
        cls.get_organisation_configuration().build_configuration(cfg)
        cls.get_subscription_configuration().build_configuration(cfg)
        cls.get_payment_notice_configuration().build_configuration(cfg)

    @classmethod
    def get_identifier_configuration(cls):
        raise NotImplementedError('`get_identifier_configuration()` must be implemented.')

    @classmethod
    def get_location_type_configuration(cls):
        raise NotImplementedError('`get_location_type_configuration()` must be implemented.')

    @classmethod
    def get_marital_type_configuration(cls):
        raise NotImplementedError('`get_marital_type_configuration()` must be implemented.')

    @classmethod
    def get_issue_type_configuration(cls):
        raise NotImplementedError('`get_issue_type_configuration()` must be implemented.')

    @classmethod
    def get_claim_configuration(cls):
        raise NotImplementedError('`get_claim_configuration()` must be implemented.')

    @classmethod
    def get_coverage_eligibility_configuration(cls):
        raise NotImplementedError('`get_coverage_eligibility_configuration()` must be implemented.')

    @classmethod
    def get_communication_request_configuration(cls):
        raise NotImplementedError('`get_communication_request_configuration()` must be implemented.')

    @classmethod
    def get_coverage_configuration(cls):
        raise NotImplementedError('`get_coverage_configuration()` must be implemented.')

    @classmethod
    def get_organisation_configuration(cls):
        raise NotImplementedError('`get_organisation_configuration()` must be implemented.')

    @classmethod
    def get_subscription_configuration(cls):
        raise NotImplementedError('`get_subscription_configuration()` must be implemented.')

    @classmethod
    def get_payment_notice_configuration(cls):
        raise NotImplementedError('`get_payment_notice_configuration()` must be implemented.')


from api_fhir_r4.configurations.generalConfiguration import GeneralConfiguration
from api_fhir_r4.configurations.R4IdentifierConfig import R4IdentifierConfig
from api_fhir_r4.configurations.R4LocationConfig import R4LocationConfig
from api_fhir_r4.configurations.R4MaritalConfig import R4MaritalConfig
from api_fhir_r4.configurations.R4IssueTypeConfig import R4IssueTypeConfig
from api_fhir_r4.configurations.R4ClaimConfig import R4ClaimConfig
from api_fhir_r4.configurations.R4InvoiceConfig import R4InvoiceConfig
from api_fhir_r4.configurations.R4CoverageEligibilityConfiguration import R4CoverageEligibilityConfiguration
from api_fhir_r4.configurations.R4CommunicationRequestConfig import R4CommunicationRequestConfig
from api_fhir_r4.configurations.R4OrganisationConfig import R4OrganisationConfig
from api_fhir_r4.configurations.R4CoverageConfig import R4CoverageConfig
from api_fhir_r4.configurations.R4SubscriptionConfig import R4SubscriptionConfig
from api_fhir_r4.configurations.R4PaymentNoticeConfig import R4PaymentNoticeConfig
# all specific configurations have to be imported before R4ApiFhirConfig
from api_fhir_r4.configurations.R4ApiFhirConfig import R4ApiFhirConfig
from api_fhir_r4.configurations.moduleConfiguration import ModuleConfiguration



from insuree.models import Profession, Education, IdentificationType, Relation, Gender

from api_fhir_r4.configurations import GeneralConfiguration


class RelationshipMapping(object):
    relationship = {}
    @classmethod
    def load(cls):
        if cls.relationship == {}:
            cls.relationship = {
                str(relation.id): relation.relation for relation in Relation.objects.all()
            }


class EducationLevelMapping(object):
    education_level = {}
    
    @classmethod
    def load(cls):
        if cls.education_level ==  {}:
            cls.education_level = {
                str(education.id): education.education for education in Education.objects.all()
                }


class PatientProfessionMapping(object):
    patient_profession = {}
    @classmethod
    def load(cls):
        if cls.patient_profession ==  {}:
            cls.patient_profession = {
                str(profession.id): profession.profession for profession in Profession.objects.all()
            }


class IdentificationTypeMapping(object):
    identification_type = {}
    @classmethod
    def load(cls):
        if cls.identification_type == {}:
            cls.identification_type = {
                identification.code: identification.identification_type for identification in IdentificationType.objects.all()
            }


class MaritalStatusMapping(object):
    marital_status = {
        "M": "Married",
        "S": "Single",
        "D": "Divorced",
        "W": "Widowed",
        "UNK": "unknown"
    }


class PatientCategoryMapping(object):
    GENDER_SYSTEM = "http://hl7.org/fhir/administrative-gender"
    AGE_SYSTEM = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/usage-context-age-type"
    imis_gender_mapping = {}
    fhir_patient_category_coding = {
        "male": {
            "system": GENDER_SYSTEM,
            "code": "male",
            "display": "Male",
        },
        "female": {
            "system": GENDER_SYSTEM,
            "code": "female",
            "display": "Female",
        },
        "adult": {
            "system": AGE_SYSTEM,
            "code": "adult",
            "display": "Adult",
        },
        "child": {
            "system": AGE_SYSTEM,
            "code": "child",
            "display": "Child",
        },
    }
    @classmethod
    def get_genders(cls):
        imis_gender_mapping = {
            # FHIR Gender code to IMIS object
            'male': Gender.objects.get(pk='M'),
            'female': Gender.objects.get(pk='F'),
        }

        o = Gender.objects.filter(pk='O').first()
        if o:
            imis_gender_mapping['other'] = o
        return imis_gender_mapping
    @classmethod
    def load(cls):
        if cls.imis_gender_mapping ==  {}:
            cls.imis_gender_mapping = cls.get_genders()

    imis_patient_category_flags = {
        "male": 1,
        "female": 2,
        "adult": 4,
        "child": 8,
    }



from .status_mapping import PaymentNoticeStatusMapping
from .payment_detail_status_mapping import PaymentNoticePaymentDetailStatusMapping
from .payment_status_mapping import PaymentNoticePaymentStatusMapping



from abc import ABC, abstractmethod
from copy import copy
from dataclasses import dataclass

from typing import Type, Dict, Callable, Any

from django.db import models

from api_fhir_r4.containedResources.containedResourceHandler import ContainedResourceManager
from api_fhir_r4.containedResources.converters import FHIRContainedResourceConverter, IMISContainedResourceConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


@dataclass
class ContainedResourceDefinition:
    # Based on FHIRContainedResourceConverter definition
    imis_field: str = None
    fhir_field: str = None
    extraction_method: Callable[[models.Model, str], Any] = lambda model, attribute: model.__getattribute__(attribute)


class AbstractContainedResourceCollection(ABC):
    def __init__(self, *args, **kwargs):
        self._build_contained_resource_managers(*args, **kwargs)

    def get_contained(self) -> Dict[BaseFHIRSerializer, ContainedResourceManager]:
        return self.__contained

    def update_reference_type(self, reference_type):
        for contained in self.__contained.values():
            if contained.imis_converter:
                contained.imis_converter.reference_type = reference_type
            if contained.fhir_converter:
                contained.fhir_converter.reference_type = reference_type

    def _build_contained_resource_managers(self, *args, **kwargs):
        # sys.setrecursionlimit(10_000)
        self.__contained = {}
        for serializer, definition in self._definitions_for_serializers().items():
            args_copy, kwargs_copy = copy(args), copy(kwargs)
            serializer_instance = serializer(*args_copy, **kwargs_copy)
            contained_manager = self._build_resource_from_serializer(serializer_instance)
            self.__contained[serializer_instance] = contained_manager

    @classmethod
    @abstractmethod
    def _definitions_for_serializers(cls) -> Dict[Type[BaseFHIRSerializer], ContainedResourceDefinition]:
        """
        Binds contained resource definition to serializer.
        """
        pass

    @classmethod
    def _build_resource_from_serializer(cls, serializer: BaseFHIRSerializer) -> ContainedResourceManager:
        reference_type = serializer.reference_type
        converter = serializer.fhirConverter
        definition = cls._definitions_for_serializers()[type(serializer)]

        fhir_converter, imis_converter = None, None

        if definition.imis_field:
            fhir_converter = FHIRContainedResourceConverter(
                imis_resource_name=definition.imis_field,
                converter=converter,
                resource_extract_method=definition.extraction_method,
                reference_type=reference_type
            )

        if definition.fhir_field:
            imis_converter = IMISContainedResourceConverter(
                resource_reference_type=definition.fhir_field,
                converter=converter,
                reference_type=reference_type
            )

        return ContainedResourceManager(fhir_converter, imis_converter, serializer, cls._create_alias(definition))

    @classmethod
    def _create_alias(cls, contained_resource_definition):
        return F"{contained_resource_definition.imis_field}__{contained_resource_definition.fhir_field}"



# Generated by Django 3.0.14 on 2022-03-13 16:07

import core.datetimes.ad_datetime
import core.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('api_fhir_r4', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='historicalsubscription',
            options={'get_latest_by': 'history_date', 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical subscription'},
        ),
        migrations.CreateModel(
            name='SubscriptionNotificationResult',
            fields=[
                ('id', models.UUIDField(db_column='UUID', default=None, editable=False, primary_key=True, serialize=False)),
                ('notified_successfully', models.BooleanField()),
                ('notification_time', core.fields.DateTimeField(db_column='Expiring', default=core.datetimes.ad_datetime.AdDatetime.now)),
                ('error', models.TextField(default=None, null=True)),
                ('subscription', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notifications_sent', to='api_fhir_r4.Subscription')),
            ],
            options={
                'db_table': 'tblSubscriptionNotificationResult',
                'managed': True,
            },
        ),
    ]



import core
from fhir.resources.R4B.activitydefinition import ActivityDefinition
from fhir.resources.R4B.codeableconcept import CodeableConcept
from fhir.resources.R4B.coding import Coding
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.money import Money
from fhir.resources.R4B.timing import Timing, TimingRepeat
from fhir.resources.R4B.usagecontext import UsageContext
from medical.models import Service

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.mapping.activityDefinitionMapping import ServiceTypeMapping, UseContextMapping, WorkflowMapping, \
    VenueMapping
from api_fhir_r4.mapping.patientMapping import PatientCategoryMapping
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.tests.mixin import FhirConverterTestMixin
from api_fhir_r4.utils import TimeUtils


class ActivityDefinitionTestMixin(GenericTestMixin, FhirConverterTestMixin):
    _TEST_SERVICE_UUID = "1234-1234-1234"
    _TEST_SERVICE_CODE = "TEST"
    _TEST_SERVICE_NAME = "Test Service"
    _TEST_SERVICE_TYPE = Service.TYPE_CURATIVE
    _TEST_SERVICE_FREQUENCY = 1
    _TEST_SERVICE_CATEGORY = Service.CATEGORY_ANTENATAL
    _TEST_SERVICE_CARE_TYPE = Service.CARE_TYPE_BOTH
    _TEST_SERVICE_PATIENT_CATEGORY = 15  # 1 & 2 & 4 & 8 (all patient categories)
    _TEST_SERVICE_PRICE = 1.0
    _TEST_SERVICE_VALIDITY_FROM = TimeUtils.str_to_date("2020-01-01", "18:50:00")
    _TEST_SERVICE_LEVEL = "D"

    _TEST_ACTIVITY_DEFINITION_RESOURCE_TYPE = "ActivityDefinition"
    _TEST_ACTIVITY_DEFINITION_STATUS = "active"
    _TEST_ACTIVITY_DEFINITION_DATE = "2020-01-01T18:50:00"
    _TEST_ACTIVITY_DEFINITION_PRICE_EXTENSION_URL = \
        f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/unit-price"
    _TEST_ACTIVITY_DEFINITION_CURRENCY = core.currency
    _TEST_ACTIVITY_DEFINITION_TIMING_FREQUENCY = 1
    _TEST_ACTIVITY_DEFINITION_TIMING_UNIT = "d"

    _TEST_ACTIVITY_DEFINITION_LEVEL_EXT_URL = \
        F"{GeneralConfiguration.get_system_base_url()}StructureDefinition/activity-definition-level"
    _TEST_ACTIVITY_DEFINITION_LEVEL_CODING_TEXT = F"Day of service"
    _TEST_ACTIVITY_DEFINITION_LEVEL_CODING_SYSTEM = \
        F"{GeneralConfiguration.get_system_base_url()}ValueSet/activity-definition-level"

    def create_test_imis_instance(self):
        imis_service = Service()
        imis_service.uuid = self._TEST_SERVICE_UUID
        imis_service.code = self._TEST_SERVICE_CODE
        imis_service.name = self._TEST_SERVICE_NAME
        imis_service.type = self._TEST_SERVICE_TYPE
        imis_service.frequency = self._TEST_SERVICE_FREQUENCY
        imis_service.category = self._TEST_SERVICE_CATEGORY
        imis_service.care_type = self._TEST_SERVICE_CARE_TYPE
        imis_service.patient_category = self._TEST_SERVICE_PATIENT_CATEGORY
        imis_service.price = self._TEST_SERVICE_PRICE
        imis_service.validity_from = self._TEST_SERVICE_VALIDITY_FROM
        imis_service.level = self._TEST_SERVICE_LEVEL
        return imis_service

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_SERVICE_UUID, imis_obj.uuid)
        self.assertEqual(self._TEST_SERVICE_CODE, imis_obj.code)
        self.assertEqual(self._TEST_SERVICE_NAME, imis_obj.name)
        self.assertEqual(self._TEST_SERVICE_TYPE, imis_obj.type)
        self.assertEqual(self._TEST_SERVICE_FREQUENCY, imis_obj.frequency)
        self.assertEqual(self._TEST_SERVICE_CATEGORY, imis_obj.category)
        self.assertEqual(self._TEST_SERVICE_CARE_TYPE, imis_obj.care_type)
        self.assertEqual(self._TEST_SERVICE_PATIENT_CATEGORY, imis_obj.patient_category)
        self.assertAlmostEqual(self._TEST_SERVICE_PRICE, imis_obj.price, places=2)
        self.assertEqual(self._TEST_SERVICE_VALIDITY_FROM, imis_obj.validity_from)
        self.assertEqual(self._TEST_SERVICE_LEVEL, imis_obj.level)

    def create_test_fhir_instance(self):
        PatientCategoryMapping.load()
        fhir_activity_definition = ActivityDefinition.construct()
        fhir_activity_definition.status = self._TEST_ACTIVITY_DEFINITION_STATUS
        fhir_activity_definition.resource_type = self._TEST_ACTIVITY_DEFINITION_RESOURCE_TYPE

        uuid_identifier = Identifier.construct()
        uuid_identifier.type = CodeableConcept.construct()
        uuid_coding = Coding.construct()
        uuid_coding.code = R4IdentifierConfig.get_fhir_uuid_type_code()
        uuid_coding.system = R4IdentifierConfig.get_fhir_identifier_type_system()
        uuid_identifier.type.coding = [uuid_coding]
        uuid_identifier.value = self._TEST_SERVICE_UUID

        sc_identifier = Identifier.construct()
        sc_identifier.type = CodeableConcept.construct()
        sc_coding = Coding.construct()
        sc_coding.code = R4IdentifierConfig.get_fhir_generic_type_code()
        sc_coding.system = R4IdentifierConfig.get_fhir_identifier_type_system()
        sc_identifier.type.coding = [sc_coding]
        sc_identifier.value = self._TEST_SERVICE_CODE

        fhir_activity_definition.identifier = [uuid_identifier, sc_identifier]

        fhir_activity_definition.date = self._TEST_ACTIVITY_DEFINITION_DATE
        fhir_activity_definition.id = self._TEST_SERVICE_UUID
        fhir_activity_definition.name = self._TEST_SERVICE_CODE
        fhir_activity_definition.title = self._TEST_SERVICE_NAME

        price_extension = Extension.construct()
        price_extension.url = self._TEST_ACTIVITY_DEFINITION_PRICE_EXTENSION_URL
        price_extension.valueMoney = Money.construct()
        price_extension.valueMoney.value = self._TEST_SERVICE_PRICE
        price_extension.valueMoney.currency = self._TEST_ACTIVITY_DEFINITION_CURRENCY

        level_extension = Extension(
                url=self._TEST_ACTIVITY_DEFINITION_LEVEL_EXT_URL,
                valueCodeableConcept=CodeableConcept(
                    coding=[Coding(
                        system=self._TEST_ACTIVITY_DEFINITION_LEVEL_CODING_SYSTEM,
                        code=self._TEST_SERVICE_LEVEL,
                        display=self._TEST_ACTIVITY_DEFINITION_LEVEL_CODING_TEXT
                    )],
                    text=self._TEST_ACTIVITY_DEFINITION_LEVEL_CODING_TEXT
                )
            )
        fhir_activity_definition.extension = [price_extension, level_extension]

        topic_codeable_concept = CodeableConcept.construct()
        topic_coding = Coding.construct()
        topic_coding.system = ServiceTypeMapping.fhir_service_type_coding[self._TEST_SERVICE_TYPE]["system"]
        topic_coding.code = ServiceTypeMapping.fhir_service_type_coding[self._TEST_SERVICE_TYPE]["code"]
        topic_coding.display = ServiceTypeMapping.fhir_service_type_coding[self._TEST_SERVICE_TYPE]["display"]
        topic_codeable_concept.coding = [topic_coding]
        fhir_activity_definition.topic = [topic_codeable_concept]

        timing = Timing.construct()
        timing.repeat = TimingRepeat.construct()
        timing.repeat.frequency = self._TEST_ACTIVITY_DEFINITION_TIMING_FREQUENCY
        timing.repeat.period = self._TEST_SERVICE_FREQUENCY
        timing.repeat.periodUnit = self._TEST_ACTIVITY_DEFINITION_TIMING_UNIT
        fhir_activity_definition.timingTiming = timing

        gender_usage_context = UsageContext.construct()
        gender_codeable_concept = CodeableConcept.construct()
        gender_coding = Coding.construct()
        gender_coding.system = UseContextMapping.fhir_use_context_coding["gender"]["system"]
        gender_coding.code = UseContextMapping.fhir_use_context_coding["gender"]["code"]
        gender_coding.display = UseContextMapping.fhir_use_context_coding["gender"]["display"]
        male_coding = Coding.construct()
        male_coding.system = PatientCategoryMapping.fhir_patient_category_coding["male"]["system"]
        male_coding.code = PatientCategoryMapping.fhir_patient_category_coding["male"]["code"]
        male_coding.display = PatientCategoryMapping.fhir_patient_category_coding["male"]["display"]
        female_coding = Coding.construct()
        female_coding.system = PatientCategoryMapping.fhir_patient_category_coding["female"]["system"]
        female_coding.code = PatientCategoryMapping.fhir_patient_category_coding["female"]["code"]
        female_coding.display = PatientCategoryMapping.fhir_patient_category_coding["female"]["display"]
        gender_codeable_concept.coding = [male_coding, female_coding]
        gender_codeable_concept.text = "Male or Female"
        gender_usage_context.valueCodeableConcept = gender_codeable_concept
        gender_usage_context.code = gender_coding

        age_usage_context = UsageContext.construct()
        age_codeable_concept = CodeableConcept.construct()
        age_coding = Coding.construct()
        age_coding.system = UseContextMapping.fhir_use_context_coding["age"]["system"]
        age_coding.code = UseContextMapping.fhir_use_context_coding["age"]["code"]
        age_coding.display = UseContextMapping.fhir_use_context_coding["age"]["display"]
        child_coding = Coding.construct()
        child_coding.system = PatientCategoryMapping.fhir_patient_category_coding["child"]["system"]
        child_coding.code = PatientCategoryMapping.fhir_patient_category_coding["child"]["code"]
        child_coding.display = PatientCategoryMapping.fhir_patient_category_coding["child"]["display"]
        adult_coding = Coding.construct()
        adult_coding.system = PatientCategoryMapping.fhir_patient_category_coding["adult"]["system"]
        adult_coding.code = PatientCategoryMapping.fhir_patient_category_coding["adult"]["code"]
        adult_coding.display = PatientCategoryMapping.fhir_patient_category_coding["adult"]["display"]
        age_codeable_concept.coding = [adult_coding, child_coding]
        age_codeable_concept.text = "Male or Female"
        age_usage_context.valueCodeableConcept = age_codeable_concept
        age_usage_context.code = age_coding

        venue_usage_context = UsageContext.construct()
        venue_codeable_concept = CodeableConcept.construct()
        venue_coding = Coding.construct()
        venue_coding.system = UseContextMapping.fhir_use_context_coding["venue"]["system"]
        venue_coding.code = UseContextMapping.fhir_use_context_coding["venue"]["code"]
        venue_coding.display = UseContextMapping.fhir_use_context_coding["venue"]["display"]
        amb_coding = Coding.construct()
        amb_coding.system = VenueMapping.fhir_venue_coding[Service.CARE_TYPE_OUT_PATIENT]["system"]
        amb_coding.code = VenueMapping.fhir_venue_coding[Service.CARE_TYPE_OUT_PATIENT]["code"]
        amb_coding.display = VenueMapping.fhir_venue_coding[Service.CARE_TYPE_OUT_PATIENT]["display"]
        imp_coding = Coding.construct()
        imp_coding.system = VenueMapping.fhir_venue_coding[Service.CARE_TYPE_IN_PATIENT]["system"]
        imp_coding.code = VenueMapping.fhir_venue_coding[Service.CARE_TYPE_IN_PATIENT]["code"]
        imp_coding.display = VenueMapping.fhir_venue_coding[Service.CARE_TYPE_IN_PATIENT]["display"]
        venue_codeable_concept.coding = [amb_coding, imp_coding]
        venue_codeable_concept.text = "ambulatory or IMP"
        venue_usage_context.valueCodeableConcept = venue_codeable_concept
        venue_usage_context.code = venue_coding

        workflow_usage_context = UsageContext.construct()
        workflow_codeable_concept = CodeableConcept.construct()
        workflow_coding = Coding.construct()
        workflow_coding.system = UseContextMapping.fhir_use_context_coding["workflow"]["system"]
        workflow_coding.code = UseContextMapping.fhir_use_context_coding["workflow"]["code"]
        workflow_coding.display = UseContextMapping.fhir_use_context_coding["workflow"]["display"]
        antenatal_coding = Coding.construct()
        antenatal_coding.system = WorkflowMapping.fhir_workflow_coding[Service.CATEGORY_ANTENATAL]["system"]
        antenatal_coding.code = WorkflowMapping.fhir_workflow_coding[Service.CATEGORY_ANTENATAL]["code"]
        antenatal_coding.display = WorkflowMapping.fhir_workflow_coding[Service.CATEGORY_ANTENATAL]["display"]
        workflow_codeable_concept.coding = [antenatal_coding]
        workflow_codeable_concept.text = "Antenatal"
        workflow_usage_context.valueCodeableConcept = workflow_codeable_concept
        workflow_usage_context.code = workflow_coding

        fhir_activity_definition.useContext = [gender_usage_context, age_usage_context,
                                               venue_usage_context, workflow_usage_context]

        return fhir_activity_definition

    def verify_fhir_instance(self, fhir_obj):
        self.assertIs(type(fhir_obj), ActivityDefinition)
        self.assertEqual(fhir_obj.resource_type, self._TEST_ACTIVITY_DEFINITION_RESOURCE_TYPE)
        self.assertEqual(fhir_obj.status, self._TEST_ACTIVITY_DEFINITION_STATUS)
        self.assertEqual(fhir_obj.id, self._TEST_SERVICE_UUID)
        self.assertEqual(fhir_obj.date, self._TEST_SERVICE_VALIDITY_FROM)
        self.assertEqual(len(fhir_obj.extension), 2)
        price_extension = fhir_obj.extension[0]
        self.assertIs(type(price_extension), Extension)
        self.assertEqual(price_extension.url, self._TEST_ACTIVITY_DEFINITION_PRICE_EXTENSION_URL)
        price_money = price_extension.valueMoney
        self.assertIs(type(price_money), Money)
        self.assertEqual(price_money.currency, self._TEST_ACTIVITY_DEFINITION_CURRENCY)
        self.assertAlmostEqual(price_money.value, self._TEST_SERVICE_PRICE, places=2)
        self.assertGreater(len(fhir_obj.identifier), 0)
        self.verify_fhir_identifier(fhir_obj, R4IdentifierConfig.get_fhir_uuid_type_code(), self._TEST_SERVICE_UUID)
        self.verify_fhir_identifier(fhir_obj, R4IdentifierConfig.get_fhir_generic_type_code(), self._TEST_SERVICE_CODE)
        timing_repeat = fhir_obj.timingTiming.repeat
        self.assertEqual(timing_repeat.frequency, self._TEST_ACTIVITY_DEFINITION_TIMING_FREQUENCY)
        self.assertEqual(timing_repeat.periodUnit, self._TEST_ACTIVITY_DEFINITION_TIMING_UNIT)
        self.assertEqual(timing_repeat.period, self._TEST_SERVICE_FREQUENCY)
        self.verify_fhir_use_context(fhir_obj, UseContextMapping.fhir_use_context_coding["gender"],
                                     [PatientCategoryMapping.fhir_patient_category_coding["male"],
                                      PatientCategoryMapping.fhir_patient_category_coding["female"]])
        self.verify_fhir_use_context(fhir_obj, UseContextMapping.fhir_use_context_coding["age"],
                                     [PatientCategoryMapping.fhir_patient_category_coding["adult"],
                                      PatientCategoryMapping.fhir_patient_category_coding["child"]])
        self.verify_fhir_use_context(fhir_obj, UseContextMapping.fhir_use_context_coding["workflow"],
                                     [WorkflowMapping.fhir_workflow_coding[self._TEST_SERVICE_CATEGORY]])
        self.verify_fhir_use_context(fhir_obj, UseContextMapping.fhir_use_context_coding["venue"],
                                     [VenueMapping.fhir_venue_coding[Service.CARE_TYPE_IN_PATIENT],
                                      VenueMapping.fhir_venue_coding[Service.CARE_TYPE_OUT_PATIENT]])

    def verify_fhir_use_context(self, fhir_obj, use_context_type, expected_use_contexts):
        if len(expected_use_contexts) > 0:
            use_contexts = [use_context for use_context in fhir_obj.useContext
                            if use_context.code.code == use_context_type["code"]]
            self.assertEqual(len(use_contexts), 1)
            self.assertEqual(use_contexts[0].code.display, use_context_type["display"])
            self.assertEqual(use_contexts[0].code.system, use_context_type["system"])
            codings = use_contexts[0].valueCodeableConcept.coding
            for use_context in expected_use_contexts:
                use_context_coding = [coding for coding in codings if coding.code == use_context["code"]]
                self.assertEqual(len(use_context_coding), 1)
                self.assertEqual(use_context_coding[0].display, use_context["display"])
                self.assertEqual(use_context_coding[0].system, use_context["system"])



from api_fhir_r4.converters import BaseFHIRConverter
from fhir.resources.R4B.codesystem import CodeSystem, CodeSystemConcept

from api_fhir_r4.utils import FhirUtils


class CodeSystemConverter(BaseFHIRConverter):

    @classmethod
    def to_imis_obj(cls, data, audit_user_id):
        raise NotImplementedError('`toImisObj()` not implemented.')  # pragma: no cover

    @classmethod
    def get_fhir_code_identifier_type(cls):
        raise NotImplementedError('`get_fhir_code_identifier_type()` not implemented.')  # pragma: no cover

    @classmethod
    def to_fhir_obj(cls, obj, reference_type):
        fhir_code_system = {}
        cls.build_fhir_code_system_status(fhir_code_system)
        cls.build_fhir_code_system_content(fhir_code_system)
        fhir_code_system = CodeSystem(**fhir_code_system)
        cls.build_fhir_id(fhir_code_system, obj)
        cls.build_fhir_url(fhir_code_system, obj)
        cls.build_fhir_code_system_name(fhir_code_system, obj)
        cls.build_fhir_code_system_title(fhir_code_system, obj)
        cls.build_fhir_code_system_date(fhir_code_system)
        cls.build_fhir_code_system_description(fhir_code_system, obj)
        cls.build_fhir_code_system_count(fhir_code_system, obj)
        cls.build_fhir_code_system_concept(fhir_code_system, obj)
        return fhir_code_system

    @classmethod
    def build_fhir_id(cls, fhir_code_system, obj):
        fhir_code_system.id = obj['id']

    @classmethod
    def build_fhir_url(cls, fhir_code_system, obj):
        fhir_code_system.url = obj['url']

    @classmethod
    def build_fhir_code_system_name(cls, fhir_code_system, obj):
        fhir_code_system.name = obj['name']

    @classmethod
    def build_fhir_code_system_title(cls, fhir_code_system, obj):
        fhir_code_system.title = obj['title']

    @classmethod
    def build_fhir_code_system_date(cls, fhir_code_system):
        from core.utils import TimeUtils
        fhir_code_system.date = TimeUtils.now()

    @classmethod
    def build_fhir_code_system_description(cls, fhir_code_system, obj):
        fhir_code_system.description = obj['description']

    @classmethod
    def build_fhir_code_system_status(cls, fhir_code_system):
        fhir_code_system['status'] = 'active'

    @classmethod
    def build_fhir_code_system_content(cls, fhir_code_system):
        fhir_code_system['content'] = 'complete'

    @classmethod
    def build_fhir_code_system_count(cls, fhir_code_system, obj):
        fhir_code_system.count = f'{len(obj["data"])}'

    @classmethod
    def build_fhir_code_system_concept(cls, fhir_code_system, obj):
        fhir_code_system.concept = []
        for item in obj["data"]:
            code_system_concept = CodeSystemConcept.construct()
            code_system_concept.code = FhirUtils.get_attr(item, obj["code_field"])
            code_system_concept.display = FhirUtils.get_attr(item, obj["display_field"])
            fhir_code_system.concept.append(code_system_concept)



from django.contrib.contenttypes.models import ContentType
from api_fhir_r4.converters import CodeSystemConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class CodeSystemSerializer(BaseFHIRSerializer):
    fhirConverter = CodeSystemConverter
    codeSystemFields = ['code_field', 'display_field', 'id', 'name', 'title', 'description', 'url']

    def __init__(self, *args, **kwargs):
        self.model = {}

        for field in self.codeSystemFields:
            self.model[field] = kwargs.pop(field, None)

        if 'data' in kwargs:
            self.model['data'] = kwargs.pop('data')
        elif 'model_name' in kwargs:
            content_type = ContentType.objects.get(model__iexact=kwargs.pop('model_name'))
            model_class = content_type.model_class()
            self.model['data'] = model_class.objects.all()
        else:
            self.model['data'] = {}

        super().__init__(*args, **kwargs)

    def to_representation(self, obj):
        return CodeSystemConverter.to_fhir_obj(self.model, self.reference_type).dict()



from .create import PaymentNoticeSerializerCreate
from api_fhir_r4.paymentNotice import PaymentNoticeConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class PaymentNoticeSerializer(BaseFHIRSerializer):
    fhirConverter = PaymentNoticeConverter()

    def create(self, validated_data):
        request = self.context.get("request")
        return PaymentNoticeSerializerCreate.create(validated_data, request)

    def update(self, instance, validated_data):
        pass



import copy
from api_fhir_r4.converters import CommunicationConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer
from claim.models import Claim, Feedback


class CommunicationSerializer(BaseFHIRSerializer):

    fhirConverter = CommunicationConverter

    def create(self, validated_data):
        claim = validated_data.get('claim_id')
        if Claim.objects.filter(id=claim, validity_to__isnull=True).count() == 0:
            raise FHIRException('Claim does not exist')
        if Feedback.objects.filter(claim__id=claim, validity_to__isnull=True).count() > 0:
            raise FHIRException('Feedback exists for this claim')
        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        from core import datetime
        copied_data['feedback_date'] = datetime.datetime.now()
        obj = Feedback.objects.create(**copied_data)
        imis_claim = Claim.objects.get(id=claim)
        imis_claim.feedback_status = Claim.FEEDBACK_DELIVERED
        imis_claim.feedback_available = True
        imis_claim.feedback = obj
        imis_claim.save()
        return obj



from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiReadTestMixin
from api_fhir_r4.configurations import GeneralConfiguration
from rest_framework.test import APITestCase


class PractitionerAPITests(GenericFhirAPITestMixin, FhirApiReadTestMixin, APITestCase):

    base_url = GeneralConfiguration.get_base_url()+'Practitioner/'
    _test_json_path = "/test/test_claimAdminPractitioner.json"

    def setUp(self):
        super(PractitionerAPITests, self).setUp()



from .to_fhir_converter import PaymentNoticeToFhirConverter
from .to_imis_converter import PaymentNoticeToImisConverter
from api_fhir_r4.converters import ReferenceConverterMixin


class PaymentNoticeConverter(
    PaymentNoticeToImisConverter, PaymentNoticeToFhirConverter
):

    @classmethod
    def to_imis_obj(cls, fhir_payment_notice, audit_user_id=None):
        return PaymentNoticeToImisConverter.to_imis_obj(fhir_payment_notice)

    @classmethod
    def to_fhir_obj(cls, imis_payment, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        return PaymentNoticeToFhirConverter.to_fhir_obj(imis_payment, reference_type)



from fhir.resources.R4B.reference import Reference
from fhir.resources.R4B.paymentnotice import PaymentNotice

from api_fhir_r4.configurations import (
    R4IdentifierConfig
)
from api_fhir_r4.converters import (
    BaseFHIRConverter,
    ReferenceConverterMixin
)
from api_fhir_r4.defaultConfig import DEFAULT_CFG
from api_fhir_r4.paymentNotice.mapping import (
    PaymentNoticeStatusMapping,
    PaymentNoticePaymentStatusMapping
)


class PaymentNoticeToFhirConverter(BaseFHIRConverter, ReferenceConverterMixin):
    @classmethod
    def to_fhir_obj(cls, imis_payment, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_payment_notice = {}
        cls.build_fhir_status(fhir_payment_notice, imis_payment)
        cls.build_fhir_created(fhir_payment_notice, imis_payment)
        cls.build_fhir_amount(fhir_payment_notice, imis_payment)
        cls.build_fhir_payment(fhir_payment_notice, imis_payment)
        cls.build_fhir_recipient(fhir_payment_notice)
        fhir_payment_notice = PaymentNotice(**fhir_payment_notice)
        cls.build_fhir_pk(fhir_payment_notice, imis_payment, reference_type)
        cls.build_fhir_identifiers(fhir_payment_notice, imis_payment)
        cls.build_fhir_request(fhir_payment_notice, imis_payment, reference_type)
        cls.build_fhir_payment_date(fhir_payment_notice, imis_payment)
        cls.build_fhir_payment_status(fhir_payment_notice, imis_payment)
        return fhir_payment_notice

    @classmethod
    def build_fhir_identifiers(cls, fhir_invoice, imis_invoice):
        identifiers = []
        cls.build_fhir_uuid_identifier(identifiers, imis_invoice)
        fhir_invoice.identifier = identifiers

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def build_fhir_status(cls, fhir_payment_notice, imis_payment):
        imis_invoice = cls._fetch_invoice_related_to_payment(imis_payment)
        fhir_payment_notice['status'] = PaymentNoticeStatusMapping\
            .to_fhir_status[imis_invoice.status]

    @classmethod
    def build_fhir_created(cls, fhir_payment_notice, imis_payment):
        fhir_payment_notice['created'] = f'{imis_payment.date_created}'

    @classmethod
    def build_fhir_request(cls, fhir_payment_notice, imis_payment, reference_type):
        imis_invoice = cls._fetch_invoice_related_to_payment(imis_payment)
        fhir_payment_notice.request = cls.build_fhir_resource_reference(
            imis_invoice.subject,
            type="Invoice",
            reference_type=reference_type
        )

    @classmethod
    def build_fhir_payment(cls, fhir_payment_notice, imis_payment):
        json_ext = imis_payment.json_ext
        if json_ext:
            if 'reconciliation' in json_ext:
                reconciliation_json_ext_dict = json_ext['reconciliation']
                if 'id' in reconciliation_json_ext_dict:
                    reference = Reference.construct()
                    resource_type = "PaymentReconciliation"
                    resource_id = reconciliation_json_ext_dict['id']
                    reference.reference = f'{resource_type}/{resource_id}'
                    fhir_payment_notice["payment"] = reference

    @classmethod
    def build_fhir_payment_date(cls, fhir_payment_notice, imis_payment):
        fhir_payment_notice.paymentDate = f'{imis_payment.date_payment}'

    @classmethod
    def build_fhir_amount(cls, fhir_payment_notice, imis_payment):
        fhir_payment_notice["amount"] = cls.build_fhir_money(imis_payment.amount_received)

    @classmethod
    def build_fhir_payment_status(cls, fhir_payment_notice, imis_payment):
        paymentStatus = PaymentNoticePaymentStatusMapping.\
            to_fhir_status[imis_payment.reconciliation_status]
        fhir_payment_notice.paymentStatus = cls.build_codeable_concept(
            code=paymentStatus,
            display=paymentStatus,
            system="http://terminology.hl7.org/CodeSystem/paymentstatus"
        )

    @classmethod
    def build_fhir_recipient(cls, fhir_payment_notice):
        default_insurance_organisation = DEFAULT_CFG['R4_fhir_insurance_organisation_config']
        reference = Reference.construct()
        resource_type = "Organization"
        resource_id = default_insurance_organisation['id']
        reference.reference = f'{resource_type}/{resource_id}'
        fhir_payment_notice["recipient"] = reference

    @classmethod
    def get_reference_obj_uuid(cls, imis_payment):
        return imis_payment.id

    @classmethod
    def get_reference_obj_id(cls, imis_payment):
        return imis_payment.id

    @classmethod
    def get_fhir_resource_type(cls):
        return PaymentNotice

    @classmethod
    def _fetch_invoice_related_to_payment(cls, imis_payment):
        imis_invoices = imis_payment.invoice_payments.all()
        imis_invoice = None
        for invoice in imis_invoices:
            imis_invoice = invoice
        return imis_invoice



import datetime

from asynctest import MagicMock
from django.test import TestCase

from api_fhir_r4.converters import ClaimConverter, ReferenceConverterMixin
from api_fhir_r4.models import Subscription, SubscriptionNotificationResult
from api_fhir_r4.subscriptions.notificationClient import RestSubscriptionNotificationClient, \
    SubscriberNotificationOutput
from api_fhir_r4.subscriptions.notificationManager import RestSubscriptionNotificationManager
from api_fhir_r4.tests import CommunicationTestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin


class AsyncContextManagerMock(MagicMock):

    async def __aenter__(self):
        return self.aenter

    async def __aexit__(self, *args):
        pass


class TestSubscriptionNotificationManager(CommunicationTestMixin, LogInMixin, TestCase):
    TEST_HEADERS_1 = """{"test-header": "123", "Authentication": "Bearer ABF13816"}"""
    EXPECTED_HEADER_1 = {'content-type': 'application/json', 'accept': 'application/json', 'test-header': '123', 'Authentication': 'Bearer ABF13816'}
    TEST_HEADERS_2 = """{"test-header": "123", "Authentication": "Bearer 61831FAB"}"""
    EXPECTED_HEADER_2 = {'content-type': 'application/json', 'accept': 'application/json', 'test-header': '123', 'Authentication': 'Bearer 61831FAB'}
    NOTIFICATION_CONTENT = {'notification_content': 'content'}

    def setUp(self) -> None:
        super(TestSubscriptionNotificationManager, self).setUp()
        self._test_user = self.get_or_create_user_api()
        self._test_subscriptions = self._create_test_subscriptions()
        self._test_resource = self._TEST_CLAIM
        self._test_converter = ClaimConverter()

    def test_sending_claim_notification(self):
        mocked_client = MagicMock()
        mocked_client.propagate_notifications = MagicMock()
        mocked_client.propagate_notifications.return_value = [
            SubscriberNotificationOutput(self._test_subscriptions[0], True),
            SubscriberNotificationOutput(self._test_subscriptions[1], False, {"ServerError": "Endpoint Unavailable"})
        ]
        test_manger = RestSubscriptionNotificationManager(fhir_converter=self._test_converter, client=mocked_client)
        test_manger.notify_subscribers_with_resource(self._test_resource, self._test_subscriptions)

        self._assert_mock_call(mocked_client)
        self._assert_saved_result()

    def _create_test_subscriptions(self):
        _valid_subscription = [self._create_valid(self.TEST_HEADERS_1), self._create_valid(self.TEST_HEADERS_2)]
        return _valid_subscription

    def _create_valid(self, headers):
        sub = Subscription(
            status=1, channel=0, endpoint='http://test-subscription-endpoint.io/post_uri/',
            headers=headers, expiring=datetime.datetime.now() + datetime.timedelta(days=10)
        )
        sub.save(username=self._test_user.username)
        return sub

    def _assert_mock_call(self, mocked_client):
        mocked_client.propagate_notifications.assert_called_with(
            self._test_converter.to_fhir_obj(self._test_resource, ReferenceConverterMixin.UUID_REFERENCE_TYPE),
            self._test_subscriptions
        )

    def _assert_saved_result(self):
        success = SubscriptionNotificationResult.objects.subscriber_notifications(self._test_subscriptions[0])
        failure = SubscriptionNotificationResult.objects.subscriber_notifications(self._test_subscriptions[1])
        self.assertEqual(len(success), 1)
        self.assertEqual(len(failure), 1)
        success, failure = success.first(), failure.first()
        self.assertTrue(success.notified_successfully)
        self.assertFalse(failure.notified_successfully)
        self.assertEqual(failure.error, """{'ServerError': 'Endpoint Unavailable'}""")



import json
import os

from api_fhir_r4.converters import CoverageEligibilityRequestConverter
from api_fhir_r4.tests import CoverageEligibilityRequestTestMixin
from fhir.resources.R4B.coverageeligibilityresponse import CoverageEligibilityResponse


class CoverageEligibilityRequestConverterTestCase(CoverageEligibilityRequestTestMixin):

    __TEST_COVERAGE_ELIGIBILITY_RESPONSE_JSON_PATH = "/test/test_coverageEligibilityResponse.json"
    __TEST_COVERAGE_ELIGIBILITY_REQUEST_JSON_PATH = "/test/test_coverageEligibilityRequest.json"

    def setUp(self):
        super(CoverageEligibilityRequestConverterTestCase, self).setUp()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self._test_coverage_eligibility_response_json_representation = open(dir_path
                                                                   + self.__TEST_COVERAGE_ELIGIBILITY_RESPONSE_JSON_PATH) \
            .read()
        self._test_coverage_eligibility_request_json_representation = open(dir_path
                                                                  + self.__TEST_COVERAGE_ELIGIBILITY_REQUEST_JSON_PATH) \
            .read()

    #def test_to_fhir_obj(self):
        #imis_coverage_eligibility_response = self.create_test_imis_instance()
        #fhir_coverage_eligibility_response = CoverageEligibilityRequestConverter.to_fhir_obj(imis_coverage_eligibility_response)
        #self.verify_fhir_instance(fhir_coverage_eligibility_response)

    def test_to_imis_obj(self):
        fhir_coverage_eligibility_request = self.create_test_fhir_instance()
        imis_coverage_eligibility_request = CoverageEligibilityRequestConverter.to_imis_obj(fhir_coverage_eligibility_request.dict(), None)
        self.verify_imis_instance(imis_coverage_eligibility_request)

    def test_create_object_from_json(self):
        dict_coverage_eligibility_response = json.loads(self._test_coverage_eligibility_response_json_representation)
        fhir_coverage_eligibility_response = CoverageEligibilityResponse(**dict_coverage_eligibility_response)
        self.verify_fhir_instance(fhir_coverage_eligibility_response)



from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.models.imisModelEnums import ContactPointSystem


class HealthFacilityOrganizationTypeMapping:
    LEGAL_FORM_CODE = 'D'
    LEGAL_FORM_DISPLAY = 'District organization'
    LEGAL_FORM_SYSTEM = 'CodeSystem/organization-legal-form'
    LEGAL_FORM_URL = f'{GeneralConfiguration.get_system_base_url()}StructureDefinition/organization-legal-form'

    ORGANIZATION_TYPE = 'prov'

    EMAIL_CONTACT_POINT_SYSTEM = ContactPointSystem.EMAIL
    PHONE_CONTACT_POINT_SYSTEM = ContactPointSystem.PHONE
    FAX_CONTACT_POINT_SYSTEM = ContactPointSystem.FAX

    ADDRESS_LOCATION_REFERENCE_URL = \
        f'{GeneralConfiguration.get_system_base_url()}/StructureDefinition/address-location-reference'

    CONTRACT_PURPOSE = {
        'code': 'PAYOR',
        'system': 'http://terminology.hl7.org/CodeSystem/contactentity-type'
    }

    LEVEL_DISPLAY_MAPPING = {
        'D': 'Dispensary',
        'C': 'Health Centre',
        'H': 'Hospital'
    }

    TYPE_DISPLAY_MAPPING = {
        'O': 'Out-patient',
        'I': 'In-patient',
        'B': 'Both'
    }

    LEVEL_SYSTEM = f'{GeneralConfiguration.get_system_base_url()}/CodeSystem/organization-hf-level'

    TYPE_SYSTEM = f'{GeneralConfiguration.get_system_base_url()}/CodeSystem/organization-hf-care-type'



from api_fhir_r4.configurations import GeneralConfiguration, R4CommunicationRequestConfig as Config
from api_fhir_r4.containedResources.converterUtils import get_from_contained_or_by_reference
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.converters.patientConverter import PatientConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.utils import DbManagerUtils
from claim.models import Claim, Feedback
from insuree.models import Insuree
from django.utils.translation import gettext as _
from fhir.resources.R4B.communication import Communication, CommunicationPayload
from fhir.resources.R4B.extension import Extension


class CommunicationConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_feedback, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_communication = {}
        cls.build_fhir_status(fhir_communication)
        fhir_communication = Communication(**fhir_communication)
        cls.build_fhir_identifiers(fhir_communication, imis_feedback)
        cls.build_fhir_subject(fhir_communication, imis_feedback, reference_type)
        cls.build_fhir_about(fhir_communication, imis_feedback, reference_type)
        cls.build_fhir_payloads(fhir_communication, imis_feedback)
        return fhir_communication

    @classmethod
    def to_imis_obj(cls, fhir_communication, audit_user_id):
        errors = []
        imis_feedback = Feedback()
        fhir_communication = Communication(**fhir_communication)
        cls._validate_fhir_status(fhir_communication)
        cls._validate_fhir_about(fhir_communication)
        cls._validate_fhir_payload(fhir_communication)
        cls.build_imis_about(imis_feedback, fhir_communication, errors)
        cls.build_imis_subject(imis_feedback, fhir_communication, errors, audit_user_id=audit_user_id)
        cls.build_imis_payloads(imis_feedback, fhir_communication, errors)
        imis_feedback.audit_user_id = audit_user_id
        cls.check_errors(errors)
        return imis_feedback

    @classmethod
    def build_fhir_identifiers(cls, fhir_communication, imis_feedback):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_feedback)
        fhir_communication.identifier = identifiers
        return fhir_communication

    @classmethod
    def build_all_identifiers(cls, identifiers, imis_object):
        # Feedback does not provide code reference
        cls.build_fhir_uuid_identifier(identifiers, imis_object)

    @classmethod
    def build_fhir_payloads(cls, fhir_communication, imis_feedback):
        fhir_communication.payload = []
        cls.build_fhir_payload(
            fhir_communication.payload,
            Config.get_fhir_care_rendered_code(),
            "yes" if imis_feedback.care_rendered is True else "no"
        )
        cls.build_fhir_payload(
            fhir_communication.payload,
            Config.get_fhir_payment_asked_code(),
            "yes" if imis_feedback.payment_asked is True else "no"
        )
        cls.build_fhir_payload(
            fhir_communication.payload,
            Config.get_fhir_drug_prescribed_code(),
            "yes" if imis_feedback.drug_prescribed is True else "no"
        )
        cls.build_fhir_payload(
            fhir_communication.payload,
            Config.get_fhir_drug_received_code(),
            "yes" if imis_feedback.drug_received is True else "no"
        )
        cls.build_fhir_payload(
            fhir_communication.payload,
            Config.get_fhir_asessment_code(),
            imis_feedback.asessment
        )

    @classmethod
    def build_fhir_payload(cls, fhir_payload, code, content_string):
        payload = {}
        payload['contentString'] = content_string
        payload = CommunicationPayload(**payload)
        payload.extension = []

        extension = Extension.construct()
        url = f'{GeneralConfiguration.get_system_base_url()}StructureDefinition/communication-payload-type'
        system = f'{GeneralConfiguration.get_system_base_url()}CodeSystem/feedback-payload'
        extension.url = url
        extension.valueCodeableConcept = cls.build_codeable_concept(
            system=system,
            code=code
        )
        payload.extension.append(extension)

        fhir_payload.append(payload)

    @classmethod
    def build_fhir_status(cls, fhir_communication):
        fhir_communication['status'] = "completed"

    @classmethod
    def build_fhir_subject(cls, fhir_communication, imis_feedback, reference_type):
        fhir_communication.subject = cls.build_fhir_resource_reference(
            imis_feedback.claim.insuree,
            reference_type=reference_type,
            type="Patient",
            display=imis_feedback.claim.insuree.chf_id
        )

    @classmethod
    def build_fhir_about(cls, fhir_communication, imis_feedback, reference_type):
        fhir_communication.about = []
        reference = cls.build_fhir_resource_reference(
            imis_feedback.claim,
            reference_type=reference_type,
            type="Claim",
            display=imis_feedback.claim.code
        )
        fhir_communication.about.append(reference)

    @classmethod
    def build_imis_about(cls, imis_feedback, fhir_communication, errors):
        claim_uuid = cls.__get_claim_reference(fhir_communication.about[0].reference)
        try:
            imis_feedback.claim = Claim.objects.get(uuid=claim_uuid, validity_to__isnull=True)
        except Exception:
            raise FHIRException(
                _('Claim does not exist')
            )

    @classmethod
    def __get_claim_reference(cls, claim):
        return claim.rsplit('/', 1)[1]

    @classmethod
    def build_imis_payloads(cls, imis_feedback, fhir_communication, errors):
        payloads = fhir_communication.payload
        for payload in payloads:
            code = cls.get_code_from_extension_codeable_concept(payload.extension[0])
            fhir_content_string = payload.contentString
            if code == Config.get_fhir_care_rendered_code():
                cls.build_imis_care_rendered(imis_feedback, fhir_content_string)
            if code == Config.get_fhir_payment_asked_code():
                cls.build_imis_payment_asked(imis_feedback, fhir_content_string)
            if code == Config.get_fhir_drug_prescribed_code():
                cls.build_imis_drug_prescribed(imis_feedback, fhir_content_string)
            if code == Config.get_fhir_drug_received_code():
                cls.build_imis_drug_received(imis_feedback, fhir_content_string)
            if code == Config.get_fhir_asessment_code():
                cls.build_imis_asessment(imis_feedback, fhir_content_string)

    @classmethod
    def build_imis_care_rendered(cls, imis_feedback, fhir_content_string):
        cls._should_be_yes_or_no(Config.get_fhir_care_rendered_code(), fhir_content_string)
        imis_feedback.care_rendered = cls._convert_bool_value(fhir_content_string)

    @classmethod
    def build_imis_payment_asked(cls, imis_feedback, fhir_content_string):
        cls._should_be_yes_or_no(Config.get_fhir_payment_asked_code(), fhir_content_string)
        imis_feedback.payment_asked = cls._convert_bool_value(fhir_content_string)

    @classmethod
    def build_imis_drug_prescribed(cls, imis_feedback, fhir_content_string):
        cls._should_be_yes_or_no(Config.get_fhir_drug_prescribed_code(), fhir_content_string)
        imis_feedback.drug_prescribed = cls._convert_bool_value(fhir_content_string)

    @classmethod
    def build_imis_drug_received(cls, imis_feedback, fhir_content_string):
        cls._should_be_yes_or_no(Config.get_fhir_drug_received_code(), fhir_content_string)
        imis_feedback.drug_received = cls._convert_bool_value(fhir_content_string)

    @classmethod
    def build_imis_asessment(cls, imis_feedback, fhir_content_string):
        imis_feedback.asessment = fhir_content_string

    @classmethod
    def _convert_bool_value(cls, fhir_content_string):
        return fhir_content_string == "yes"

    @classmethod
    def _should_be_yes_or_no(cls, code, content):
        if content not in ["yes", "no"]:
            raise ValidationError(f"Value for '{code}' must be either 'yes' or 'no' but is '{content}'")

    @classmethod
    def get_reference_obj_id(cls, imis_feedback):
        return imis_feedback.uuid

    @classmethod
    def get_fhir_resource_type(cls):
        return Communication

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        return DbManagerUtils.get_object_or_none(
            Feedback,
            **cls.get_database_query_id_parameteres_from_reference(reference))


    @classmethod
    def get_reference_obj_uuid(cls, imis_feedback):
        return imis_feedback.uuid

    @classmethod
    def _validate_fhir_about(cls, fhir_communication):
        if not fhir_communication.about:
            raise FHIRException(
                _('about is required')
            )

    @classmethod
    def build_imis_subject(cls, imis_feedback, fhir_communication, errors, audit_user_id):
        if not fhir_communication.subject:
            raise FHIRException(
                _('subject is required')
            )
        else:
            insuree = get_from_contained_or_by_reference(
                fhir_communication.subject, None , PatientConverter, audit_user_id)
            if insuree:
                imis_feedback.claim.insuree = insuree
                imis_feedback.claim.insuree.chf_id = insuree.chf_id
            cls.valid_condition(not imis_feedback.claim.insuree, _('Missing or invalid `subject` reference'), errors)

    @classmethod
    def _validate_fhir_status(cls, fhir_communication):
        if not fhir_communication.status:
            raise FHIRException(
                _('status is required field')
            )
        else:
            if not fhir_communication.status == 'completed':
                raise FHIRException(
                    _('status value must be = completed')
                )

    @classmethod
    def _validate_fhir_payload(cls, fhir_communication):
        if not fhir_communication.payload:
            raise FHIRException(
                _('payload is required')
            )
        else:
            if len(fhir_communication.payload) != 5:
                raise FHIRException(
                    _('payload must have 5 elements')
                )
            else:
                required_payload = [
                    'CareRendered', 'PaymentAsked',
                    'DrugPrescribed', 'DrugReceived', 'Asessment'
                ]
                extension_values = []
                for payload in fhir_communication.payload:
                    extension_values.append(
                        cls.get_code_from_extension_codeable_concept(payload.extension[0])
                    )
                required_payload.sort()
                extension_values.sort()
                if required_payload != extension_values:
                    raise FHIRException(
                        _(F"payload must support all five elemets:"
                          " 'CareRendered', 'PaymentAsked',"
                          " 'DrugPrescribed', 'DrugReceived', 'Asessment'")
                    )
